"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
import logging
import threading
import time
from queue import Queue
from enum import Enum
import os
import json
import pandas as pd
from PyQt5.QtCore import pyqtSignal, QSize, QTimer, pyqtSlot
from PyQt5.QtGui import QFont, QTextCursor, QColor
from PyQt5.QtWidgets import QMainWindow, QLabel, QVBoxLayout, QPushButton, QPlainTextEdit, QWidget
import pyqtgraph as pg
from tkinter import Tk, ttk, Toplevel, Label

from wiliot_tools.utils.wiliot_gui.wiliot_gui import WiliotGui, popup_message, FONT_BUTTONS
from wiliot_testers.wiliot_tester_tag_test import load_test_suite_file, TesterName
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api
from wiliot_testers.offline.modules.offline_utils import ConfigDefaults, open_session, save_screen
from wiliot_testers.offline.modules.offline_utils import printing_test_window, printing_sgtin_window
from wiliot_testers.offline.modules.offline_utils import get_validation_data_for_scanning
from wiliot_testers.tester_utils import open_json_cache, upload_conclusion

from wiliot_testers.offline.modules.offline_tag_testing import TagTesting, TagStates
from wiliot_testers.offline.modules.offline_printing_and_validation import PrintingAndValidation, ValidationStates
from wiliot_testers.offline.configs.moh_instruction import moh_instruction, lang_map
from wiliot_testers.offline.configs.global_vars_and_enums import MainDefaults, ASTRIX

from wiliot_testers.offline.modules.offline_log import DataLogging


class ErrorCode(Enum):
    Main = 100
    TagTesting = 200
    PrintingAndValidation = 300


class HwErrors(Enum):
    R2R = 10
    PRINTER = 20
    SCANNER = 30
    ATTENUATOR = 50
    SENSOR = 60


class MohType(Enum):
    MISSING_LABEL = 'Missing Label'
    STOP_CRITERIA = 'stop Criteria'
    VALIDATION_FAILED = 'Validation Failed'
    FILES_ERROR = 'Files Error'
    PRINTER_ERROR = 'Printer Error'
    SCANNER_ERROR = 'Scanner Error'
    ARDUINO_ERROR = 'Arduino Error'
    NO_PRINTER_ACK = 'Missed Printing'
    GENERAL_ERROR = 'General Error'


class MainStates(Enum):
    STOP_CRITERIA = 1
    UPDATE_PLOT = 2
    RESTART_AFTER_MOH = 3


class TestParametersException(Exception):
    def __init__(self, msg):
        self.message = msg
        super().__init__(self.message)


class Main(QMainWindow):
    sig = pyqtSignal(str)

    def __init__(self, test_config_path=None, *args, **kwargs):

        # init communication between classes
        self.test_start_event = threading.Event()
        self.test_end_event = threading.Event()
        self.validation_end_event = threading.Event()
        self.stop_app_event = threading.Event()
        self.exception_q = Queue(maxsize=MainDefaults.MAX_QUEUE_SIZE)
        self.tag_results_queue = Queue(maxsize=MainDefaults.MAX_QUEUE_SIZE)

        # init test parameters including tests suites
        self.test_parameters = TestParameters()

        # init user inputs
        self.is_gui = test_config_path is None
        self.user_config = None
        tag_print_config = self.run_main_gui(test_config_path)

        # init log
        self.test_logger = DataLogging(exception_queue=self.exception_q, log_config=self.user_config)
        self.logger = logging.getLogger(self.test_logger.get_logger_name())
        self.logger.info(f"Wiliot's package version = {get_version()}")

        # init test parameters including tests suites
        self.test_parameters.set_logger_name(logger_name=self.test_logger.get_logger_name())

        # init printer config:
        self.external_hw_config = self.test_parameters.init_external_hw_config()

        # select test suites
        self.test_parameters.set_specific_test(user_inputs=self.user_config)

        # open printing GUI
        self.print_config = self.run_printing_gui(tag_print_config)
        self.test_logger.update_log_config({**self.external_hw_config, **self.print_config})
        self.scan_validation_df = self.run_scanning_gui()

        # run application
        self.tag_testing = None
        self.print_log_validation = None
        self.tag_testing_thread = None
        self.print_log_validation_thread = None
        self.r2r_obj = None
        self.moh_obj = ManualOperationHandle(exception_queue=self.exception_q, logger_name=self.logger.name,
                                             is_gw_r2r=False)
        self.gw_version = ''
        self.run_app()

        # run GUI:
        self.tested = None
        self.text_box = None
        self.data_line = None
        self.reel_label = None
        self.update_timer = None
        self.last_tested_num = 0
        self.last_passed_num = 0
        super(Main, self).__init__(*args, **kwargs)
        self.open_ui(wiliot_ver=get_version(), gw_version=self.gw_version)

    def run_main_gui(self, test_config_path):
        tag_print_config = {}
        if self.is_gui:
            self.user_config, tag_print_config = open_session(
                test_suite_list=self.test_parameters.get_tests_suites_names())
        else:
            try:
                with open(test_config_path, 'r') as f:
                    self.user_config = json.load(f)
            except Exception as e:
                raise Exception(f'Main: init: error while running without gui: '
                                f'could not read user inputs from the following path: {test_config_path}, due to: {e}')
        return tag_print_config

    def check_conversation_label_error(self, new_reel, tag_print_config):
        need_conv_label = self.user_config['gen'].lower() == 'gen3' and self.user_config['Environment'].lower() == 'prod'
        conv_label_not_changed = not tag_print_config.get('conversion_label_changed', False)
        not_allow_same_label = not self.user_config['sameConversionLabel']
        need_warning = new_reel and need_conv_label and conv_label_not_changed and not_allow_same_label
        return need_warning

    def run_printing_gui(self, tag_print_config=None):
        tag_print_config = tag_print_config if tag_print_config is not None else {}
        is_valid = True
        msg = ''
        if self.user_config['toPrint'] == 'Yes' and not tag_print_config.get('run_printing_gui', False):
            new_tag_print_config = {}
            if self.user_config['printingFormat'] == 'Test':
                new_tag_print_config, is_valid = printing_test_window()
                if not is_valid:
                    msg = 'Impossible printing values entered by the user, the program will exit now'

            elif self.user_config['printingFormat'] in MainDefaults.PRINTING_GUI_TYPE:
                new_tag_print_config, is_valid, new_reel = \
                    printing_sgtin_window(env=self.user_config['Environment'],
                                          owner_id=self.user_config['OwnerId'],
                                          printing_format=self.user_config['printingFormat'],
                                          gen=self.user_config['gen'])
                if not is_valid:
                    msg = 'user exited the program'

                if self.check_conversation_label_error(new_reel=new_reel, tag_print_config=tag_print_config):
                    msg += '\nConversion label was not updated although NEW reel was assembled'
                    is_valid = False

            else:
                msg = 'user chose unsupported printing format!!!'

            for k, v in new_tag_print_config.items():
                tag_print_config[k] = v

        if not is_valid:
            raise Exception(f'Main: run_printing_gui: {msg}')

        return tag_print_config

    def run_scanning_gui(self):
        df = pd.DataFrame()
        if self.user_config['toPrint'].lower() == 'no' and self.user_config['QRRead'].lower() == 'yes':
            df = get_validation_data_for_scanning()
        return df

    def run_app(self):
        # init all classes
        self.tag_testing = TagTesting(test_start_event=self.test_start_event,
                                      test_end_event=self.test_end_event,
                                      validation_end_event=self.validation_end_event,
                                      stop_app_event=self.stop_app_event,
                                      moh_object=self.moh_obj,
                                      exception_queue=self.exception_q,
                                      tag_results_queue=self.tag_results_queue,
                                      log_obj=self.test_logger,
                                      test_object=self.test_parameters,
                                      user_config=self.user_config,
                                      hw_config=self.external_hw_config)
        if not self.tag_testing.is_init:
            err = self.moh_obj.get_all_exceptions()
            raise Exception(f'Got the following errors during TagTesting init:\n{err}')
        self.moh_obj.set_is_gw_r2r(self.tag_testing.is_r2r)
        self.print_log_validation = PrintingAndValidation(test_start_event=self.test_start_event,
                                                          test_end_event=self.test_end_event,
                                                          validation_end_event=self.validation_end_event,
                                                          stop_app_event=self.stop_app_event,
                                                          moh_object=self.moh_obj,
                                                          exception_queue=self.exception_q,
                                                          tag_results_queue=self.tag_results_queue,
                                                          log_obj=self.test_logger,
                                                          test_object=self.test_parameters,
                                                          user_config=self.user_config,
                                                          gw_version=self.tag_testing.get_gw_version(),
                                                          get_duplicated_tags=self.tag_testing.get_all_duplicated_tags,
                                                          scanning_data=self.scan_validation_df,
                                                          get_gw_signal=self.tag_testing.get_gw_gpio_signal)
        if not self.print_log_validation.is_init:
            err = self.moh_obj.get_all_exceptions()
            raise Exception(f'Got the following errors during PrintingAndValidation init:\n{err}')
        # set r2r counter:
        self.r2r_obj = self.tag_testing.r2r if self.tag_testing.is_r2r else self.print_log_validation.r2r
        self.set_r2r_configuration()

        # open threads for all classes
        self.tag_testing_thread = threading.Thread(target=self.tag_testing.run)
        self.print_log_validation_thread = threading.Thread(target=self.print_log_validation.run)

        self.tag_testing_thread.start()
        self.print_log_validation_thread.start()

        self.gw_version = self.tag_testing.get_gw_version()

    def set_r2r_configuration(self):
        if self.tag_testing.is_r2r:
            self.print_log_validation.set_r2r_counter_func(self.tag_testing.r2r.get_counter)
            # update instruction strings
            for lan in moh_instruction[MohType.ARDUINO_ERROR.value].keys():
                moh_instruction[MohType.ARDUINO_ERROR.value][lan] = \
                    [s.replace('Arduino', 'Gateway') for s in moh_instruction[MohType.ARDUINO_ERROR.value][lan]]
            self.logger.info('Gateway working instead of Arduino')

    @pyqtSlot(str)
    def append_debug(self, string):
        self.text_box.appendPlainText(string)  # +'\n')

    def open_ui(self, wiliot_ver, gw_version):
        """
        opens the run main GUI that will present the run data and gives to the user ability to Stop/Continue/Pause
        """
        stop_label = QLabel("If you want to stop this run, press stop")
        stop_label.setFont(QFont(MainDefaults.FONT_NAME, MainDefaults.FONT_SIZE))
        self.reel_label = QLabel("Reel Name: ")
        self.reel_label.setFont(QFont(MainDefaults.FONT_NAME, MainDefaults.FONT_SIZE))
        self.reel_label.setStyleSheet('.QLabel {padding-top: 10px; font-weight: bold; font-size: 25px; color:#ff5e5e;}')
        self.reel_label.setFont(QFont(MainDefaults.FONT_NAME, MainDefaults.FONT_SIZE))
        self.tested = QLabel("Tested = 0, Passed = 0, Yield = -1%")
        self.tested.setFont(QFont(MainDefaults.FONT_NAME, MainDefaults.FONT_SIZE))
        last_tag_str = QLabel("Last Tag Passed: ")
        last_tag_str.setFont(QFont(MainDefaults.FONT_NAME, MainDefaults.FONT_SIZE, weight=QFont.Bold))
        layout = QVBoxLayout()

        stop = QPushButton("Stop")
        stop.setStyleSheet("background-color: #FD4B4B")
        stop.setFont(QFont(MainDefaults.FONT_NAME, MainDefaults.FONT_SIZE))
        stop.setFixedSize(QSize(300, 22))
        stop.pressed.connect(self.stop_app_event.set)

        c = ConsolePanelHandlerGUI(self.sig)
        c.setLevel(logging.WARNING)
        self.text_box = QPlainTextEdit()
        self.text_box.setPlaceholderText("Warnings will be printed here")
        self.text_box.setMaximumBlockCount(1000)
        self.text_box.setReadOnly(True)

        self.logger.addHandler(c)
        self.sig.connect(self.append_debug)
        self.text_box.moveCursor(QTextCursor.End)

        graphWidget = pg.PlotWidget()
        graphWidget.setBackground('w')
        # Add Title
        graphWidget.setTitle("Yield over time", color=QColor("56C2FF"), size="20pt")
        styles = {"color": "#f00", "font-size": "14px"}
        graphWidget.setLabel("left", "Yield for the last 50 tags [%]", **styles)
        graphWidget.setLabel("bottom",
                             f"Last tag location [x*{MainDefaults.CALCULATE_INTERVAL}+{MainDefaults.CALCULATE_ON}]",
                             **styles)
        pen = pg.mkPen(color=(255, 0, 0))
        self.data_line = graphWidget.plot([], [], pen=pen)
        versions = QLabel("PyWiliot Version: {}\nGW Version: {}".format(wiliot_ver, gw_version))
        versions.setFont(QFont(MainDefaults.FONT_NAME, MainDefaults.FONT_SIZE, weight=QFont.Bold))

        layout.addWidget(self.reel_label)
        layout.addWidget(stop_label)
        layout.addWidget(stop)
        layout.addWidget(last_tag_str)
        layout.addWidget(self.tested)
        # layout.addWidget(self.debug)
        layout.addWidget(self.text_box)
        layout.addWidget(graphWidget)
        layout.addWidget(versions)

        w = QWidget()
        w.setLayout(layout)
        self.setCentralWidget(w)
        self.show()

        # updates the GUI and stops all if exception happened
        self.update_timer = QTimer()
        self.update_timer.setInterval(MainDefaults.GUI_UPDATE_TIME)
        self.update_timer.timeout.connect(self.gui_run)
        self.update_timer.start()

    def check_stop_criteria(self, tested, responded, passed, is_passed_list, yield_over_time):
        # check Stop Conditions:
        need_to_stop = False
        messages = []
        if not self.user_config['ignoreStop']:

            # check max fail in a row:
            max_failed = int(self.user_config['maxFailStop'])
            if len(is_passed_list) > max_failed:
                max_fail_in_a_row = not any(is_passed_list[-max_failed:])
                if max_fail_in_a_row:
                    msg = str('*' * 100) + f'\nThere are {self.user_config["maxFailStop"]} fails in a row, ' \
                                           f'Run will stop\n' + str('*' * 100)
                    self.logger.warning(msg)
                    messages.append('max failed in a row happened')
                    need_to_stop = True

            # check yield drop:
            if len(yield_over_time) >= MainDefaults.N_YIELD_SAMPLES_UNDER_THRESHOLD:
                yield_drop_happened = True
                for ii in range(1, MainDefaults.N_YIELD_SAMPLES_UNDER_THRESHOLD):
                    if yield_over_time[-ii] > int(self.user_config['maxYieldStop']):
                        yield_drop_happened = False
                        break
                if yield_drop_happened:
                    msg = str('*' * 100) + f'\nThe yield-over-time of the last tags is below ' \
                                           f'{self.user_config["maxYieldStop"]}%, Run will stop\n' + str('*' * 100)
                    self.logger.warning(msg)
                    messages.append('yield drop happened')
                    need_to_stop = True

            # check responding vs pass:
            high_responded_vs_passed = \
                responded > int(self.user_config['pass_response_diff_offset']) and \
                (responded - passed) / tested > int(self.user_config['pass_response_diff']) / 100
            if high_responded_vs_passed:
                msg = str('*' * 100) + f'\nThere is a significant differences between passed yield ' \
                                       f'and responded yield:{self.user_config["pass_response_diff"]}%, ' \
                                       f'Run will stop\n' + str('*' * 100)
                self.logger.warning(msg)
                messages.append('High difference between passed and responding happened')
                need_to_stop = True

        return need_to_stop, messages

    def timed_popup(self):
        params_dict = {
            'space': {'value': '', 'widget_type': 'label'},
            'msg': {'value': 'Please wait until the Post-Run-Validation is completed\n',
                    'widget_type': 'label', 'options': {'font': ("Gudea", 20, 'bold')}},
            'progress': {'value': '', 'widget_type': 'label'}
        }
        wg = WiliotGui(params_dict=params_dict,
                       do_button_config=False, parent=None, exit_sys_upon_cancel=False)
        max_length = self.print_log_validation.get_validation_q_maxsize()

        def update_progress():
            current_length = self.print_log_validation.get_validation_q_size()
            progress = round(100 * (1 - (current_length / max_length)), 2)
            wg.update_widget('progress', f'{progress}%')

            if current_length <= 0:
                wg.exit_gui()

        update_progress()
        wg.add_recurrent_function(cycle_ms=1000, function=update_progress)
        wg.run()

    def wait_for_tag_test_end(self):
        t_start = time.time()
        while self.tag_testing.is_test_running() and \
                time.time() - t_start <= self.test_parameters.get_total_test_time():
            time.sleep(0.200)

    def check_end_of_test(self, tested, passed):
        # check end of test according to number of tested:
        if tested >= int(self.user_config['desiredTags']):
            self.wait_for_tag_test_end()
            msg = f'---------------------------Desired tags have reached ({tested})---------------------------'
            self.logger.warning(msg)
            return True

        # check end of test according to number of passed:
        if passed >= int(self.user_config['desiredPass']):
            self.wait_for_tag_test_end()
            msg = f'---------------------------Desired passed tags have reached ({passed})---------------------------'
            self.logger.warning(msg)
            return True
        return False

    @staticmethod
    def calculate_yield_over_time(is_passed_list):
        yield_over_time = \
            [sum(is_passed_list[x - MainDefaults.CALCULATE_ON:x]) / MainDefaults.CALCULATE_ON * 100
             for x in range(MainDefaults.CALCULATE_ON, len(is_passed_list), MainDefaults.CALCULATE_INTERVAL)]
        return yield_over_time

    def gui_run(self):

        try:

            if self.stop_app_event.is_set():
                self.exit()
                return

            tested = self.tag_testing.get_tested() - int(self.tag_testing.is_test_running())
            passed = self.tag_testing.get_passed()
            is_passed_list = self.tag_testing.get_is_passed_list()
            responded = self.tag_testing.get_responded()
            missing_labels = self.tag_testing.get_missing_label_count()
            self.reel_label.setText("Reel Name: " + self.test_logger.get_reel_name())

            if tested == 0:
                passed_yield = -1.0
                responded_yield = -1
            else:
                passed_yield = passed / tested * 100
                responded_yield = responded / tested * 100
            if self.test_logger.get_reel_name() == '':
                raise ValueError('REEL NAME PROBLEM')
            self.tested.setText(f'Tested = {tested}, Passed = {passed}, '
                                f'Responded Yield = {round(responded_yield, 2)}%, '
                                f'Yield = {round(passed_yield, 2)}%\n Tag reel location:'
                                f' {self.print_log_validation.get_tag_reel_location() + 1},'
                                f' Missing labels = {missing_labels}')

            # update plot:
            yield_over_time = self.calculate_yield_over_time(is_passed_list)
            x = list(range(len(yield_over_time)))
            self.data_line.setData(x, yield_over_time)

            # check end of test:
            end_of_test = None if self.user_config['ignoreStop'] else self.check_end_of_test(tested=tested,
                                                                                             passed=passed)
            if end_of_test:
                self.exit()
                return

            # check Stop Conditions:
            need_to_stop, messages = self.check_stop_criteria(tested=tested, responded=responded, passed=passed,
                                                              is_passed_list=is_passed_list,
                                                              yield_over_time=yield_over_time)
            if need_to_stop:
                for msg in messages:
                    self.exception_q.put(f'Main Exception: run: state{MainStates.STOP_CRITERIA.name}: {msg}',
                                         block=False)
        except Exception as e:
            self.exception_q.put(f'Main Exception: run: state{MainStates.UPDATE_PLOT.name}: {e}', block=False)
        # check manual operation handling:
        exceptions_status = self.moh_obj.handle_exceptions()
        while exceptions_status['pause_run']:
            self.moh_obj.set_moh_is_needed(True)
            # open moh GUI:
            self.update_timer.stop()
            is_end_of_run = False
            continue_run = False
            if exceptions_status['moh_type'] == MohType.MISSING_LABEL:
                is_end_of_run = self.end_of_run_pop_up() if self.is_gui else True
            if not is_end_of_run:
                continue_run = \
                    self.moh_obj.moh_pop_up(
                        moh_type=exceptions_status['moh_type'],
                        error_code=exceptions_status['error_code'],
                        msg=exceptions_status['exception_str'],
                        scan_again_func=self.print_log_validation.update_scanning_results_after_exception
                        if self.print_log_validation.is_scanner else None)
            self.restart_run_after_moh(continue_run)
            exceptions_status = self.moh_obj.handle_exceptions()

            if not continue_run:
                self.exit()
                break
        self.update_state_machines()
        self.moh_obj.set_moh_is_needed(False)
        self.update_timer.start()

    def restart_run_after_moh(self, continue_run):
        self.logger.info('Start the run after MOH')
        is_print = self.user_config['toPrint'].lower() == 'yes'
        is_scanner = self.user_config['QRRead'].lower() == 'yes'
        is_r2r = self.user_config['isR2R'].lower() == 'yes'
        try:
            self.tag_testing.gw_reset_and_config()
            if is_r2r:
                self.r2r_obj.reconnect()
                self.r2r_obj.send_start_to_r2r()
                if not continue_run:
                    self.r2r_obj.end_of_run()
            if is_print:
                self.print_log_validation.restart_printer_after_moh()
            if is_scanner:
                self.print_log_validation.scanner.reconnect()

        except Exception as e:
            self.exception_q.put(f'Main Exception: run: state{MainStates.RESTART_AFTER_MOH.name}: {e}', block=False)

    def update_state_machines(self):
        # to re-do WAIT_END, since results are not yet ready:
        if self.print_log_validation.last_state == ValidationStates.WAIT_END and \
                self.tag_testing.last_state in [TagStates.TAG_TEST, TagStates.PROCESS_TEST]:
            self.print_log_validation.last_state = ValidationStates.SET_NEXT_PRINT
        return

    @staticmethod
    def end_of_run_pop_up():
        params_dict = {
            'space': {'value': '', 'widget_type': 'label'},
            'is_new': {'value': '         End of Run?    \n',
                       'widget_type': 'label', 'options': {'font': ("Gudea", 20, 'bold')}}}
        wg = WiliotGui(params_dict=params_dict,
                       do_button_config=False, parent=None, exit_sys_upon_cancel=False)
        wg.button_configs(submit_button_text='Yes', cancel_button_text='No')
        out = wg.run()
        return out is not None

    def upload_data_to_cloud_and_summary(self):
        # upload data to cloud and present summary
        tested = self.tag_testing.get_tested()
        passed = self.tag_testing.get_passed()
        responded = self.tag_testing.get_responded()
        is_passed_list = self.tag_testing.get_is_passed_list()
        missing_labels = self.tag_testing.get_missing_label_count()

        if tested == 0:
            return

        if not self.is_gui and 'upload_to_cloud' in self.user_config.keys():
            values = {'upload': self.user_config['upload_to_cloud'], 'comments': ''}
        else:
            ttfp_list = self.tag_testing.get_ttfp_list()
            ttfp_avg = self.tag_testing.calculating_ttfp_avg(ttfp_list)
            values = save_screen(tested=tested,
                                 passed=passed,
                                 responded=responded,
                                 yield_=passed / tested * 100 if tested > 0 else -1,
                                 missing_labels=self.tag_testing.get_missing_label_count(),
                                 ttfgp_avg=ttfp_avg,
                                 ttfp_max_error=MainDefaults.MAX_TTFP_ERROR)

        if values['upload'].lower() == 'yes':
            file_size = os.stat(self.test_logger.get_packets_data_path()).st_size
            if file_size < MainDefaults.MAX_FILE_SIZE:
                try:
                    res = upload_to_cloud_api(batch_name=self.test_logger.get_reel_name(),
                                              tester_type='offline-test',
                                              run_data_csv_name=self.test_logger.get_run_data_path(),
                                              packets_data_csv_name=self.test_logger.get_packets_data_path(),
                                              env=self.user_config['Environment'],
                                              owner_id=self.user_config['OwnerId'],
                                              logger_=self.logger.name, is_path=True)
                except Exception as e:
                    self.logger.warning(f'Main: upload_data_to_cloud_and_summary: {e}')
                    res = False

                if self.is_gui:
                    upload_conclusion(succeeded_csv_uploads=res)
            else:
                e_msg = 'Test files are too large, please upload using manual upload app'
                popup_message(msg=e_msg, bg='red')
                self.logger.warning(f'Main: upload_data_to_cloud_and_summary: {e_msg}')
        else:
            self.logger.info('Data was not Uploaded to cloud, due to user selection')

        # summary message:
        yield_over_time = self.calculate_yield_over_time(is_passed_list)
        responded_yield = responded / tested * 100 if tested > 0 else -1
        passed_yield = passed / tested * 100 if tested > 0 else -1
        msg = f"Stopped by the operator.\nReels yield_over_time is: |{yield_over_time}" \
              f"| interval: |{MainDefaults.CALCULATE_INTERVAL}|, on: |{MainDefaults.CALCULATE_ON}\n" \
              f"User comments: {values['comments']}\n" \
              f"Tested = {tested}, Passed = {passed}, " \
              f"Responded Yield = {responded_yield}%, Yield = {passed_yield}%, Missing labels = {missing_labels}"
        self.logger.info(msg)

    def stop_app(self):
        if self.tag_testing_thread.is_alive():
            self.logger.info('Waiting for Tag Testing thread to stop all processes...')
            self.tag_testing_thread.join()
            self.logger.info('Tag Testing thread was done')
        if self.print_log_validation_thread.is_alive():
            self.logger.info('Waiting for Printing Log and Validation thread to stop all processes...')
            self.print_log_validation_thread.join()
            self.logger.info('Printing Log and Validation thread was done')

    def exit(self):
        self.stop_app_event.set()
        try:
            offset_size = self.print_log_validation.get_validation_q_maxsize()
            if offset_size > 1:
                self.timed_popup()
            self.stop_app()
            if self.is_gui:
                self.moh_obj.get_all_exceptions()
            self.upload_data_to_cloud_and_summary()
        except Exception as e:
            self.logger.warning(f'Main: exit: Got Exception during upload data: [{e}]\n'
                                f'please make sure data was uploaded')
        self.update_timer.stop()
        self.close()


class TestParameters(object):
    def __init__(self, logger_name=None):
        self.logger = None
        if logger_name is not None:
            self.logger = logging.getLogger(logger_name)
        # set config folder
        app_dir = os.path.abspath(os.path.dirname(__file__))
        self.dir_config = os.path.join(app_dir, 'configs')
        # set test suite:
        self.tests_suite = None
        self.all_tests_suites = load_test_suite_file(TesterName.OFFLINE.value)
        self.total_time = 0
        self.user_inputs = None
        self.stop_criteria = None

    def get_tests_suites_names(self):
        return list(self.all_tests_suites.keys())

    def set_logger_name(self, logger_name):
        self.logger = logging.getLogger(logger_name)

    def set_specific_test(self, user_inputs):
        self.user_inputs = user_inputs
        self.set_test_suite()
        calculated_time = self.calculate_total_test_time()
        self.set_total_test_time(total_time=calculated_time)

        self.stop_criteria = self.set_stop_criteria()

    def set_stop_criteria(self):
        stop_criteria = {}
        if 'pass_response_diff' in self.user_inputs.keys():
            max_diff_btwn_pass_and_response = {}
            try:
                max_diff_btwn_pass_and_response['value'] = int(self.user_inputs['pass_response_diff'])
            except Exception as e:
                raise TestParametersException(f'set_stop_criteria: Could not convert the pass_response_diff to number:'
                                              f' {e}')
            if 'pass_response_diff_offset' in self.user_inputs.keys():
                try:
                    max_diff_btwn_pass_and_response['offset'] = int(self.user_inputs['pass_response_diff_offset'])
                except Exception as e:
                    raise TestParametersException(f'set_stop_criteria: Could not convert the pass_response_diff_offset '
                                                  f'to number: {e}')
            stop_criteria['max_diff_btwn_pass_and_response'] = max_diff_btwn_pass_and_response

        fields = {'desiredTags': 'max_tested_tags', 'desiredPass': 'max_pass_tags'}
        for f_user, f_stop in fields.items():
            if f_user in self.user_inputs.keys():
                stop_criteria[f_stop] = self.user_inputs[f_user]
            else:
                raise TestParametersException(f'set_stop_criteria: could not find {f_user} in '
                                              f'the user inputs: {self.user_inputs.keys()}')
        return stop_criteria

    def calculate_total_test_time(self):
        total_time = 0
        for stage in self.tests_suite['tests']:
            if 'maxTime' in stage.keys():
                total_time += stage['maxTime']
            if 'delayBeforeNextTest' in stage.keys():
                total_time += stage['delayBeforeNextTest']
            if 'DelayAfterTest' in stage.keys():
                total_time += stage['DelayAfterTest']
        total_time *= int(self.tests_suite.get('n_cycles', 1))
        return total_time

    def set_total_test_time(self, total_time):
        self.total_time = total_time

    def get_total_test_time(self):
        return self.total_time

    def set_test_suite(self):
        # read file:
        if self.user_inputs['testName'] in self.all_tests_suites.keys():
            self.tests_suite = self.all_tests_suites[self.user_inputs['testName']]
        else:
            raise TestParametersException(f'set_test_suite: specified test, {self.user_inputs["testName"]}, '
                                          f'is not part of test suit file: {self.all_tests_suites.keys()}')

    def get_selected_test_suite(self):
        return self.tests_suite

    def init_external_hw_config(self):
        config_defaults = ConfigDefaults()
        printer_config_path = os.path.join(self.dir_config, 'configs_for_printer_values.json')
        printer_config_values = open_json_cache(self.dir_config, printer_config_path,
                                                config_defaults.get_printer_defaults())
        if 'printingFormat' in printer_config_values.keys():
            del printer_config_values['printingFormat']
        external_hw_config_path = os.path.join(self.dir_config, 'test_configs.json')
        external_hw_config_values = open_json_cache(self.dir_config, external_hw_config_path,
                                                    config_defaults.get_external_hw_defaults())
        return {**printer_config_values, **external_hw_config_values}


class ManualOperationHandle(object):
    def __init__(self, exception_queue, logger_name, is_gw_r2r):
        """

        @param exception_queue:
        @type exception_queue: Queue
        """
        self.exception_q = exception_queue
        self.logger = logging.getLogger(logger_name)
        self.is_moh_needed = False
        self.is_gw_r2r = is_gw_r2r
        self.critical_moh_types = []
        self.set_critical_errors()

    def set_critical_errors(self):
        self.critical_moh_types = [MohType.GENERAL_ERROR, MohType.FILES_ERROR, MohType.STOP_CRITERIA,
                                   MohType.NO_PRINTER_ACK]

    def get_manual_operation_is_needed(self):
        return self.is_moh_needed or not self.exception_q.empty()

    def set_moh_is_needed(self, is_needed):
        self.is_moh_needed = is_needed

    def set_is_gw_r2r(self, is_r2r):
        self.is_gw_r2r = is_r2r

    @staticmethod
    def error_code_to_moh(error_code):
        if compare_error_code(error_code, ErrorCode.TagTesting.value, 0) and \
                compare_error_code(error_code, TagStates.GET_TRIGGER.value, 2):
            moh_type = MohType.MISSING_LABEL
        elif compare_error_code(error_code, ErrorCode.Main.value, 0) and \
                compare_error_code(error_code, MainStates.STOP_CRITERIA.value, 2):
            moh_type = MohType.STOP_CRITERIA
        elif compare_error_code(error_code, ErrorCode.PrintingAndValidation.value, 0) and \
                compare_error_code(error_code, ValidationStates.VALIDATION.value, 2):
            moh_type = MohType.VALIDATION_FAILED
        elif compare_error_code(error_code, ErrorCode.PrintingAndValidation.value, 0) and \
                compare_error_code(error_code, ValidationStates.PRINTER_ACK.value, 2):
            moh_type = MohType.NO_PRINTER_ACK
        elif compare_error_code(error_code, ErrorCode.PrintingAndValidation.value, 0) and \
                compare_error_code(error_code, ValidationStates.LOG_DATA.value, 2):
            moh_type = MohType.FILES_ERROR
        elif compare_error_code(error_code, HwErrors.PRINTER.value, 1):
            moh_type = MohType.PRINTER_ERROR
        elif compare_error_code(error_code, HwErrors.SCANNER.value, 1):
            moh_type = MohType.SCANNER_ERROR
        elif compare_error_code(error_code, HwErrors.R2R.value, 1):
            moh_type = MohType.ARDUINO_ERROR
        else:  # any error will pop up a msg
            moh_type = MohType.GENERAL_ERROR
        return moh_type

    def get_all_exceptions(self):

        def show_exceptions(exceptions):
            instructions_str = moh_instruction[MohType.GENERAL_ERROR.value][MainDefaults.DEFAULT_LANGUAGE]
            param_dict_ex = {f'exc{i}': {'value': exc, 'widget_type': 'label', 'columnspan': 3}
                             for i, exc in enumerate(exceptions)}
            param_dict_ins = {
                'space': {'value': '', 'widget_type': 'label'},
                'language': {'text': 'select language', 'value': [lang_code for lang_code in lang_map.keys()]},
                '-INSTRUCTIONS-': {'value': '\n'.join(instructions_str), 'widget_type': 'label', 'columnspan': 3,
                                   'options': {'font': FONT_BUTTONS}},
                'quit': [{'space': {'value': '', 'widget_type': 'label'}},
                         {'quit': {'value': 'Quit', 'widget_type': 'button', 'options': {'bg': 'white', 'fg': 'red'}}}]
            }
            param_dict = {**param_dict_ex, **param_dict_ins}
            wg = WiliotGui(params_dict=param_dict, do_button_config=False, exit_sys_upon_cancel=False, theme='warning')

            wg.add_event(widget_key='language',
                         event_type='<<ComboboxSelected>>',  # need to add args for this event_type
                         command=lambda args: self.update_instruction(app_in=wg, moh_type=MohType.GENERAL_ERROR))
            self.update_instruction(app_in=wg, moh_type=MohType.GENERAL_ERROR)
            wg.add_event(widget_key='quit_quit', event_type='button', command=wg.on_cancel)
            wg.run()

        n_exceptions = self.exception_q.qsize()
        err = []
        for _ in range(n_exceptions):
            exception_str = self.exception_q.get(block=False)
            err.append(exception_str)
            self.logger.warning(exception_str)
        if len(err) > 0:
            show_exceptions(err)

        return err

    def handle_exceptions(self):
        exceptions_status = {'pause_run': False}
        n_exceptions = self.exception_q.qsize()
        for _ in range(n_exceptions):
            exception_str = self.exception_q.get(block=False)
            error_code = get_error_code(exception_str)
            self.logger.warning(f'{ASTRIX} ErrorCode:{error_code} - {exception_str} {ASTRIX}')

            moh_type = self.error_code_to_moh(error_code=error_code)
            exceptions_status = {'pause_run': True, 'exception_str': exception_str,
                                 'error_code': error_code, 'moh_type': moh_type}
            return exceptions_status

        return exceptions_status

    def moh_pop_up(self, error_code, moh_type, msg, scan_again_func=None):
        language_code = MainDefaults.DEFAULT_LANGUAGE  # Default to English
        instructions_str = moh_instruction[moh_type.value][language_code] \
            if moh_type.value in moh_instruction and language_code in moh_instruction[moh_type.value] \
            else ['error instructions are not available']

        moh_name = moh_type.value
        nice_msg = msg.replace(': ', '\n')
        if self.is_gw_r2r:
            moh_name = moh_name.replace('Arduino', 'Gateway')
            nice_msg = nice_msg.replace('Arduino', 'Gateway')

        param_dict = {
            'title': {'value': 'Manual Handling is Needed!', 'widget_type': 'label',
                      'options': {'font': ("Gudea", 18, "bold")}, 'columnspan': 3},
            'moh_name': {'value': moh_name, 'widget_type': 'label', 'options': {'font': FONT_BUTTONS}, 'columnspan': 3},
            'instructions_title': {'value': 'Instructions:', 'widget_type': 'label', 'columnspan': 3,
                                   'options': {'font': ("Gudea", 12, "underline")}},
            'language': {'text': 'select language', 'value': [lang_code for lang_code in lang_map.keys()]},
            '-INSTRUCTIONS-': {'value': '\n'.join(instructions_str), 'widget_type': 'label', 'columnspan': 3,
                               'options': {'font': FONT_BUTTONS}},
            'error_code': {'value': f'ErrorCode:{error_code}', 'widget_type': 'label', 'columnspan': 3},
            'msg': {'value': f'{nice_msg}', 'widget_type': 'label', 'columnspan': 3},
        }

        if self.must_end_run(moh_type):
            param_dict['stop'] = [{'space': {'value': '', 'widget_type': 'label'}},
                                  {'stop': {'value': 'Stop', 'widget_type': 'button'}}]
        elif moh_type == MohType.VALIDATION_FAILED and scan_again_func is not None:
            param_dict['rescan'] = {'value': 'Scan Again', 'widget_type': 'button'}
            param_dict['scan_output'] = {'value': '', 'widget_type': 'label'}

        wg = WiliotGui(params_dict=param_dict, do_button_config=False, exit_sys_upon_cancel=False, theme='warning',
                       height_offset=10, width_offset=0)

        wg.add_event(widget_key='language',
                     event_type='<<ComboboxSelected>>',  # need to add args for this event_type
                     command=lambda args: self.update_instruction(app_in=wg, moh_type=moh_type))
        self.update_instruction(app_in=wg, moh_type=moh_type)
        if self.must_end_run(moh_type):
            wg.add_event(widget_key='stop_stop', event_type='button', command=wg.on_cancel)
        else:
            if 'rescan' in param_dict:
                wg.add_event(widget_key='rescan', event_type='button',
                             command=lambda: wg.update_widget(widget_key='scan_output', new_value=scan_again_func()))
            wg.button_configs(submit_button_text='Continue', cancel_button_text='Stop')

        values = wg.run()
        return values is not None

    @staticmethod
    def update_instruction(app_in, moh_type):
        values = app_in.get_all_values()
        language_code = lang_map[values['language']]  # Update the language code
        new_instructions = moh_instruction[moh_type.value][language_code] \
            if moh_type.value in moh_instruction and language_code in moh_instruction[moh_type.value] \
            else ['error instructions are not available']
        app_in.update_widget(widget_key='-INSTRUCTIONS-', new_value='\n'.join(new_instructions))

        if values['language'] == 'Hebrew':
            app_in.widgets['-INSTRUCTIONS-'].configure(anchor="e", justify='right')
        else:
            app_in.widgets['-INSTRUCTIONS-'].configure(anchor="w", justify='left')

    def must_end_run(self, moh_type):
        return moh_type in self.critical_moh_types


def get_error_code(msg):
    error_code = 0
    for m in ErrorCode.__members__:
        if f'{m} Exception'.lower() in msg.lower():
            error_code += ErrorCode[m].value
            break

    for h in HwErrors.__members__:
        if f'{h}:'.lower() in msg.lower():
            error_code += HwErrors[h].value
            break

    if 'run:' in msg.lower():
        if compare_error_code(error_code, ErrorCode.Main.value, 0):
            states = MainStates
        elif compare_error_code(error_code, ErrorCode.TagTesting.value, 0):
            states = TagStates
        elif compare_error_code(error_code, ErrorCode.PrintingAndValidation.value, 0):
            states = ValidationStates
        else:
            return error_code

        for m in states:
            if f'state{m.name}'.lower() in msg.lower():
                error_code += m.value
                break

    return error_code


def compare_error_code(error_code1, error_code2, char_num):
    error_code1_str = str(error_code1).zfill(3)
    error_code2_str = str(error_code2).zfill(3)
    return error_code1_str[char_num] == error_code2_str[char_num]


class ConsolePanelHandlerGUI(logging.Handler):

    def __init__(self, sig):
        logging.Handler.__init__(self)
        self.stream = sig

    def handle(self, record):
        rv = self.filter(record)
        if rv:
            self.acquire()
            try:
                self.emit(record)
            finally:
                self.release()
        return rv

    def emit(self, record):
        try:
            self.stream.emit(self.format(record))
        except RecursionError:
            raise
        except Exception as e:
            self.handleError(record)


if __name__ == '__main__':
    import argparse
    from PyQt5.QtWidgets import QApplication

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--file_name', action='store', dest='file_name',
                        help='file_name - Test Configuration File Name (Json Format, *.json)')
    args = vars(parser.parse_args())
    config_file_name = args['file_name']

    app = QApplication([])
    window = Main(test_config_path=config_file_name)
    app.exec_()



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import logging
import os
from wiliot_api import ManufacturingClient, TesterType
from wiliot_core import GetApiKey


def upload_to_cloud_api(batch_name, tester_type, run_data_csv_name=None, packets_data_csv_name=None,
                        env='', is_batch_name_inside_logs_folder=True, logger_=None, is_path=False,
                        client=None, owner_id='wiliot-ops'):
    """
    uploads a tester log to Wiliot cloud
    :type batch_name: string
    :param batch_name: folder name of the relevant log
    :type run_data_csv_name: string
    :param run_data_csv_name: name of desired run_data log to upload,
                              should contain 'run_data' and end with .csv
    :type packets_data_csv_name: string
    :param packets_data_csv_name: name of desired packets_data log to upload,
                               should contain 'packets_data' and end with .csv
    :type tester_type: string or TesterType
    :param tester_type: name of the tester the run was made on (offline, tal15k, conversion, yield)
    :type env: string (prod, dev, test)
    :param env: to what cloud environment should we upload the files
    :type is_batch_name_inside_logs_folder: bool
    :param is_batch_name_inside_logs_folder: flag to indicate if the batch_name is the regular run folder (logs) or
                                             this function is being used in a way we will need the full path
    :return: True for successful upload, False otherwise
    """
    # Logger setup
    if logger_ is None:
        logger = logging.getLogger()
    else:
        logger = logging.getLogger(logger_)

    # Check tester type
    if not isinstance(tester_type, TesterType):
        is_tester_type_valid = len([t for t in TesterType if t.value == tester_type])
        if is_tester_type_valid:
            tester_type = TesterType(tester_type)
        else:
            logger.warning('Unsupported tester_type inserted to upload_to_cloud_api()\nPlease change it and retry')
            return False

    # Check file names
    if run_data_csv_name and 'run_data' not in run_data_csv_name:
        logger.warning('Unsupported run_data_csv_name inserted to upload_to_cloud_api()\nPlease change it and retry')
        return False
    if packets_data_csv_name and 'packets_data' not in packets_data_csv_name:
        logger.warning('Unsupported packets_data_csv_name inserted to upload_to_cloud_api()\nPlease change it and retry')
        return False

    # Environment validation
    if env == 'production' or env == '':
        env = 'prod'
    env = env.strip('/')
    if env not in {'prod', 'test', 'dev'}:
        logger.warning(f'Unsupported env value was inserted (env = {env})')
        return False

    # Check user credentials and client setup
    g = GetApiKey(gui_type='ttk', env=env, owner_id=owner_id)
    api_key = g.get_api_key()
    if not api_key:
        logger.warning('could not extract user credentials. please check warnings')
        return False
    client = ManufacturingClient(api_key=api_key, env=env, logger_=logger.name) if client is None else client

    # Upload process
    run_upload_status = False
    packet_upload_status = False

    try:
        if run_data_csv_name is not None:
            run_data_file_path = os.path.join("logs" if is_batch_name_inside_logs_folder
                                              else "", batch_name, run_data_csv_name) if not is_path \
                                              else run_data_csv_name
            run_upload_status = client.upload_testers_data(tester_type=tester_type,
                                                           file_path=run_data_file_path)
    except Exception as e:
        logger.warning(f'A problem occurred in upload_to_cloud_api: {e}')

    try:
        if packets_data_csv_name is not None:
            packets_data_file_path = os.path.join("logs" if is_batch_name_inside_logs_folder
                                                  else "", batch_name, packets_data_csv_name) if not is_path \
                                                  else packets_data_csv_name
            packet_upload_status = client.upload_testers_data(tester_type=tester_type,
                                                              file_path=packets_data_file_path)
    except Exception as e:
        logger.warning(f'A problem occurred in upload_to_cloud_api: {e}')

    merged_status = run_upload_status and packet_upload_status
    if merged_status:
        logger.info('\n-----------------------------------------------------------------------\n'
                    'upload to cloud is finished successfully\n'
                    '-----------------------------------------------------------------------')
    return merged_status



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

# here is a list of parameters which are change based on the packet flow version.
# the flow min and flow mac MUST be ordered

packet_parameters = {
    'packet_type': [
        {'flow_min': None, 'flow_max': '0x503', 'value': 0},  # version <= 2.3
        {'flow_min': '0x504', 'flow_max': '0x5ff', 'value': 1},  # version 2.4, 2.5, 2.9
        {'flow_min': '0x600', 'flow_max': None, 'value': 3},  # version 3.0
    ],
}


def get_packet_param(flow_in, param_name):
    """
    This function extract the value of the specified parameter name based on the packet flow version
    :param flow_in the packet flow version
    :type flow_in str
    :param param_name the flow parameter
    :type param_name str
    """
    flow_in = flow_in.lower().replace('.', '')
    flow_in = flow_in if flow_in.startswith('0x') else '0x' + flow_in
    if param_name not in packet_parameters.keys():
        raise Exception(f'get_flow_params: param_name is not in the map: {packet_parameters.keys()}')

    param_values = packet_parameters[param_name]
    for packet_flow in param_values:
        if packet_flow['flow_max'] and flow_in > packet_flow['flow_max']:
            continue
        if packet_flow['flow_min'] and flow_in < packet_flow['flow_min']:
            raise Exception(f'flow_in is {flow_in} but we could not find a match flow')
        return packet_flow['value']



import PySimpleGUI as sg
import threading
from wiliot_testers.utils.ppfp_tool import DataPullGUI
from wiliot_testers.partners.simple_test.simple_test import SimpleTest

OWNER_ID = ''


def custom_progress_bar(window, title, total_time, shared_status):
    layout = [[sg.Text(title, key='-STATUS-')],
              [sg.ProgressBar(total_time, orientation='h', size=(20, 20), key='-PROGRESS-')],
              [sg.Cancel()]]

    progress_window = sg.Window('Progress', layout, keep_on_top=True)
    for i in range(total_time):
        if ((shared_status.get('test_complete', False) and title == 'Processing')
                or (shared_status.get('upload_complete', False) and title == 'Uploading')):
            break
        event, _ = progress_window.read(timeout=1000)
        if event in (sg.WIN_CLOSED, 'Cancel'):
            break
        progress_window['-PROGRESS-'].update_bar(i + 1)

    progress_window.close()


def run_simple_test(owner_id, test_name, output_path, window, test_time, shared_status):
    try:
        test = SimpleTest(owner_id=owner_id,
                          test_name=test_name,
                          output_path=output_path,
                          test_time=test_time,
                          tester_type='offline',
                          test_env='prod')
    except Exception:
        sg.popup_error('Problem initializing the test')
        return
    shared_status['test'] = test
    shared_status['test_complete'] = True
    window.write_event_value('-TEST COMPLETE-', None)


def upload_to_cloud(window, shared_status):
    if shared_status.get('test'):
        shared_status['test'].cloud_upload()
        shared_status['upload_complete'] = True
        window.write_event_value('-UPLOAD COMPLETE-', None)


def create_gui():
    sg.theme('GreenTan')

    left_layout = [
        [sg.Text('Test and cloud upload')],
        [sg.Text('Test Name'), sg.InputText(key='test_name', default_text='SimpleTest', size=(15, 1))],
        [sg.Text('Time to Test'), sg.InputText(key='test_time', default_text='10', size=(5, 1))],
        [sg.Text('Output')],
        [sg.InputText(size=(20, 1)), sg.FolderBrowse(key='target_dir')],
        [sg.Button('Start Test')],
        [sg.Text('Upload data to cloud', visible=False, key='-UPLOAD TEXT-')],
        [sg.Button('Upload', visible=False, key='-UPLOAD BUTTON-')]]

    right_layout = [
        [sg.Text('Cloud data pull', size=(20, 1))],
        [sg.Text('Test Name', visible=False), sg.InputText(key='cloud_test_name', visible=False)],
        [sg.Text('Output Directory', visible=False), sg.InputText(key='cloud_output_dir', visible=False)],
        [sg.Button('Get Results')]]

    layout = [
        [sg.Column(left_layout), sg.VSeparator(), sg.Column(right_layout)]]

    window = sg.Window('GUI', layout, size=(450, 250))

    shared_status = {'test_complete': False, 'test': None, 'upload_complete': False}

    while True:
        event, values = window.read()
        if event == sg.WIN_CLOSED:
            break

        elif event == 'Start Test':
            run_and_upload = threading.Thread(target=run_simple_test,

                                              args=(OWNER_ID, values['test_name'], values['target_dir'], window,
                                                    values['test_time'], shared_status),

                                              daemon=True)
            run_and_upload.start()
            custom_progress_bar(window, 'Processing', int(values['test_time']) + 10, shared_status)

        elif event == '-TEST COMPLETE-':
            window['-UPLOAD TEXT-'].update(visible=True)
            window['-UPLOAD BUTTON-'].update(visible=True)

        elif event == '-UPLOAD BUTTON-':
            threading.Thread(target=upload_to_cloud, args=(window, shared_status), daemon=True).start()
            custom_progress_bar(window, 'Uploading', 10, shared_status)

        elif event == '-UPLOAD COMPLETE-':
            common_run_name = shared_status['test'].get_common_run_name()
            out_dir = shared_status['test'].get_out_put_dir()
            window['cloud_test_name'].update(value=common_run_name, visible=True)
            window['cloud_output_dir'].update(value=values['target_dir'], visible=True)

        elif event == 'Get Results':
            gui = DataPullGUI(owner_id=OWNER_ID, single_crn=values['cloud_test_name'],
                              output_dir=values['cloud_output_dir'])
            gui.run()

    window.close()


if __name__ == '__main__':
    create_gui()



"""This file is only for yield because the listener log writes only the first packet of each adva
so from here we can all the packets a tag transmitted."""

from os.path import isfile
import re
from wiliot_core import PacketList, Packet
import tkinter as tk
from tkinter import filedialog


def browse_window():
    root = tk.Tk()
    root.withdraw()
    log_path = filedialog.askopenfilename(title="Select a LOG file", filetypes=[("LOG files", "*.log")])

    if log_path:
        packet_list = lines_to_packet_list(log_path)
        csv_path = log_path[:-4] + '.csv'
        packet_list.to_csv(csv_path)
        print('File saved')
    else:
        print("No file selected")


def lines_to_packet_list(log_path=None):
    packet_list = PacketList()
    seen_advas = set()
    if log_path is None:
        print('no log path was found. Export csv failed')
        return
    try:
        if isfile(log_path):
            f = open(log_path, 'r')
            lines = f.readlines()
            string_value = "'raw': 'process_packet(\""
            for line in lines:
                if string_value in line or 'is_valid_tag_packet:True' in line:
                    if string_value in line:
                        re_match = re.search("process_packet\(\"(\w+)\"", line)
                        if re_match:
                            packet_raw = str(re_match.groups()[0])
                        else:
                            print("Pattern did not match!")
                            continue
                    else:
                        re_match = re.search(",packet:(\w+)", line)
                        packet_raw = str(re_match.groups(1)[0])

                    if 'time_from_start:' in line:
                        re_match = re.search("time_from_start:(\d+.\d+)", line)
                        packet_time = float(re_match.groups(1)[0])
                    else:
                        re_match = re.search(r"'time': (\d+)\.(\d+),", line)
                        if re_match:
                            packet_time = float(re_match.group(1))
                        else:
                            continue
                    current_packet = Packet(raw_packet=packet_raw, time_from_start=packet_time)
                    if current_packet.packet_data['adv_address'] not in seen_advas:
                        packet_list.append(current_packet)
                        seen_advas.add(current_packet.packet_data['adv_address'])
            f.close()
        return packet_list
    except Exception as e:
        print('export packets from log was failed due to: {}'.format(e))
        return None, None


if __name__ == '__main__':
    browse_window()




#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import os
import time
from PIL import Image
from wiliot_core import GetApiKey, WiliotDir, open_json
from wiliot_api import ManufacturingClient
from wiliot_tools.utils.wiliot_gui.wiliot_gui import *


class DataPullGUI:
    """
    To create .exe file for this script use the next line:
    pywiliot-testers> pyinstaller --onefile --windowed --add-data "./wiliot_testers/docs/wiliot_logo.png;./docs" ./wiliot_testers/utils/ppfp_tool.py
    """

    def __init__(self, owner_id='', single_crn='', output_dir=None):
        current_script = os.path.abspath(__file__)
        current_dir = os.path.dirname(current_script)
        os.chdir(current_dir)

        self.params_dict = {
            # Owner ID Section
            'owner_id': {
                'text': 'Owner ID',
                'value': owner_id,  # Assuming owner_id is predefined
                'widget_type': 'entry',
            },

            # Environment Section
            'env': {
                'text': 'Environment',
                'value': 'Production',
                'options': ['Production', 'Test'],
                'widget_type': 'combobox',
            },

            # Tester Type Section
            'tester_type': {
                'text': 'Tester Type',
                'value': 'Offline',
                'options': ['Offline', 'Sample'],
                'widget_type': 'combobox',
            },

            # Common Run Name Insert Mode Section
            'select_mode': {
                'text': 'Select mode for Common Run Name Insert',
                'value': '',
                'widget_type': 'label',
            },
            'crn_mode': {
                'text': 'Select mode for Common Run Name Insert',
                'value': 'Single CRN',
                'widget_type': 'combobox',
                'options': ['Single CRN', 'CRN List (CSV)'],
            },

            # CRN and CSV File Section
            'crn_col': {
                'text': 'CRN',
                'value': single_crn,
                'widget_type': 'entry',
            },
            'csv_file_col': {
                'text': 'CSV File',
                'value': '',
                'widget_type': 'file_input',
            },

            # Target Directory Section
            'target_dir': {
                'text': 'select_target_directory',
                'value': output_dir if output_dir is not None else '',
                'widget_type': 'file_input',
                'options': 'folder',
            },
        }

    def run(self):

        def on_combobox_change(*args):
            selected_value = ppfp_gui.widgets_vals['crn_mode'].get()
            if selected_value == 'Single CRN':
                ppfp_gui.update_widget('crn_col', disabled=False)
                ppfp_gui.update_widget('csv_file_col', disabled=True)
            elif selected_value == 'CRN List (CSV)':
                ppfp_gui.update_widget('crn_col', disabled=True)
                ppfp_gui.update_widget('csv_file_col', disabled=False)

        ppfp_gui = WiliotGui(params_dict=self.params_dict)
        ppfp_gui.add_event(widget_key='crn_mode', command=on_combobox_change)
        on_combobox_change()
        values = ppfp_gui.run()

        try:
            tester_type = 'offline-test' if values['tester_type'] == 'Offline' else 'sample-test'
            g = GetApiKey(gui_type='ttk', env=values['env'][0:4].lower(), owner_id=values['owner_id'])
            api_key = g.get_api_key()
            if not api_key:
                raise Exception('User configuration check failed')
            client = ManufacturingClient(api_key=api_key, env=values['env'][0:4].lower())
            rsp = None

            if values['crn_mode'] == 'Single CRN':
                common_run_name_list = [values['crn_col'].strip()]
            else:
                common_run_name_list = []
                with open(values['csv_file_col'], 'r') as f:
                    for line in f:
                        common_run_name_list.append(line.strip())
            try:
                for common_run_name in common_run_name_list:
                    out_file_path = os.path.join(values['target_dir'], f'{common_run_name}.zip')
                    with open(out_file_path, 'wb') as out_file:
                        rsp = client.get_file_for_ppfp(common_run_name, tester_type, out_file)
            except Exception as e:
                print(f'problem get file from cloud due to {e}')
                popup_message(f'An error occurred while loading CSV file - please check it')

            if rsp:
                print('Job Success')
                done_layout = {
                    'done': {
                        'text': '',
                        'value': 'Job Succeeded',
                        'widget_type': 'label'
                    }
                }
                popup_message(f'Job Succeeded!\nzip file is saved under {values["target_dir"]}', title='Success')

            else:
                popup_message(f'An error occurred while getting data from cloud')

        except Exception as e:
            print(e)
            popup_message(f'An error occurred: {e}')


if __name__ == '__main__':
    gui = DataPullGUI()
    gui.run()



"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

from wiliot_testers.tester_utils import open_json_cache
from wiliot_testers.offline.modules.offline_utils import ConfigDefaults, get_printed_value, \
    get_print_user_config_file, DefaultGUIValues
from wiliot_tools.test_equipment.test_equipment import BarcodeScanner, CognexDataMan
from wiliot_testers.wiliot_tester_tag_result import FailureCodes, WiliotTesterTagResultList
from wiliot_testers.offline.configs.global_vars_and_enums import PrintingAndValidationDefaults, ASTRIX
from wiliot_testers.offline.modules.offline_r2r_controller import R2R
from wiliot_testers.offline.modules.offline_bendurance_r2r_controller import R2RBendurance

import os
import json
import time
import logging
from threading import Event
import datetime
from enum import Enum
from queue import Queue
import socket
import pandas as pd


class PrintingAndValidationException(Exception):
    def __init__(self, msg):
        self.message = msg
        super().__init__(f'PrintingAndValidation Exception: {self.message}')


class PrinterException(PrintingAndValidationException):
    def __init__(self, msg):
        super().__init__(f'Printer: {msg}')


class ScannerException(PrintingAndValidationException):
    def __init__(self, msg):
        super().__init__(f'Scanner: {msg}')


class R2RException(PrintingAndValidationException):
    def __init__(self, msg):
        super().__init__(f'R2R: {msg}')


class ValidationStates(Enum):
    IDLE = 0
    STOP = 99
    WAIT_MOH = 88
    WAIT_START = 81
    WAIT_END = 82
    START_TEST = 1
    PRE_VALIDATION = 2
    VALIDATION = 3
    PRINTER_ACK = 4
    SET_NEXT_PRINT = 5
    END_TEST = 6
    NEXT_TAG = 7
    LOG_DATA = 8
    PRINTER_CART_READY = 9


class ValidatedBin(Enum):
    # even values are good
    UNTESTED = 0  # not tested by the scanner
    PRINT_ACK_ONLY = 2  # got acknowledgement from printer
    MATCH = 4  # ext. id/black square as expected
    # odd values are bad
    MISMATCH = 3  # ex id/black square not as expected (should be a different ex.id or black square)
    NO_READ = 1  # expect to read ex id, but no success reading it
    ERROR = 99


class PrintingAndValidation(object):
    def __init__(self, test_start_event, test_end_event, validation_end_event, stop_app_event, moh_object, test_object,
                 exception_queue, tag_results_queue, log_obj=None, user_config=None, gw_version='',
                 get_duplicated_tags=None, scanning_data=None, get_gw_signal=None):
        """

        @param test_start_event: event is set when tag is under test
        @type test_start_event: Event
        @param test_end_event: event is set when tag test was completed
        @type validation_end_event: Event
        @param validation_end_event: event is set when validation ends if GW move R2R
        @type test_end_event: Event
        @param moh_object: an object contains all functions related to MOH event, Manual Operation Handling
        @type moh_object: Object
        @param test_object: an object contains all functions related to test parameters
        @type test_object: TestParameters
        @param exception_queue: queue for all exceptions, handled in the main class
        @type exception_queue: Queue
        @param tag_results_queue: queue for tag test results, sent from the tag test class
        @type tag_results_queue: Queue
        @param user_config: all user configuration inputs
        @type user_config: dict or None
        """
        self.test_start_event = test_start_event
        self.test_end_event = test_end_event
        self.validation_end_event = validation_end_event
        self.stop_app_event = stop_app_event
        self.exception_q = exception_queue
        self.tag_results_q = tag_results_queue
        self.end_of_test_tags = []
        self.moh = moh_object
        self.test = test_object
        self.logger = logging.getLogger(log_obj.get_logger_name())
        self.is_print = user_config['toPrint'].lower() == 'yes'
        self.is_scanner = user_config['QRRead'].lower() == 'yes'
        self.is_r2r = user_config['isR2R'].lower() == 'yes' and user_config['typeR2R'].lower() == 'arduino'
        self.get_duplicated_tags = get_duplicated_tags
        self.get_gw_signal = get_gw_signal
        self.change_tag_to_duplicated = []
        self.scanning_data = scanning_data
        self.printer_cart_is_ready = False

        self.last_state = ValidationStates.IDLE
        self.tag_run_location = PrintingAndValidationDefaults.FIRST_TAG_RUN_LOCATION - 1
        self.external_id_counter = 0
        self.tag_result_to_validate = None

        self.is_init = False
        self.r2r = None
        self.r2r_counter_func = None
        try:
            if self.is_r2r:
                if 'bendurance' in user_config.keys() and user_config['bendurance'].lower() == 'yes':
                    self.r2r = R2RBendurance(exception_queue=exception_queue,
                                             logger_name=log_obj.get_logger_name(),
                                             exc_obj=R2RException)
                else:
                    self.r2r = R2R(exception_queue=exception_queue,
                                   logger_name=log_obj.get_logger_name(),
                                   exc_obj=R2RException)
                self.r2r_counter_func = self.r2r.get_counter
            
            self.printer = Printer(is_print=self.is_print,
                                   exception_queue=exception_queue,
                                   logger_name=log_obj.get_logger_name(),
                                   printing_config=log_obj.get_log_config())
            if self.printer.bi_directional:
                self.init_printer_cart()

            self.scanner = Scanner(is_scanner=self.is_scanner,
                                   exception_queue=exception_queue,
                                   logger_name=log_obj.get_logger_name(),
                                   scanning_config=user_config,
                                   is_print=self.is_print)

            self.data_logging = log_obj
            self.data_logging.run_data_init(is_printing=self.is_print, gw_version=gw_version,
                                            test_suite_dict=test_object.get_selected_test_suite())

            self.is_init = True
        except Exception as e:
            if isinstance(e, PrintingAndValidationException):
                self.exception_q.put(f'init: {e}', block=False)
            else:
                self.exception_q.put(PrintingAndValidationException(f'{e}').__str__(), block=False)
            return

        # init Queues
        try:
            self.validation_q = None
            self.init_validation_q()
            self.printing_type_q = None
            self.init_printing_type_q()
        except Exception as e:
            if isinstance(e, PrintingAndValidationException):
                self.exception_q.put(f'init: queues: {e}', block=False)
            else:
                self.exception_q.put(PrintingAndValidationException(f'init: queues: {e}').__str__(), block=False)
            return

        # move to the first tag:
        try:
            if self.is_print:
                if self.printer.bi_directional:
                    gw_signals = self.get_gw_signal()
                    if gw_signals:
                        self.logger.info(f'clean the following gw signals during init: {gw_signals}')
                self.set_next_printing_type(False)
            if self.is_r2r:
                self.r2r.move_to_the_first_tag()
        except Exception as e:
            if isinstance(e, PrintingAndValidationException):
                self.exception_q.put(f'init: move to the first tag: {e}', block=False)
            else:
                self.exception_q.put(PrintingAndValidationException(f'init: move to the first tag: {e}').__str__(),
                                     block=False)
            return

    def get_r2r_counter(self):
        if self.r2r_counter_func is None:
            return 0
        return self.r2r_counter_func()

    def set_r2r_counter_func(self, r2r_counter_func=None):
        self.r2r_counter_func = r2r_counter_func

    def init_validation_q(self):
        if self.is_scanner:
            q_size = self.scanner.scanning_config['QRoffset']
        elif self.is_print:
            q_size = int(self.printer.printing_config['printOffset']) - 1 + \
                     PrintingAndValidationDefaults.QUEUE_VALIDATION_OFFSET
        else:
            q_size = 1  # No validation is needed
        self.validation_q = Queue(maxsize=int(q_size))

    def init_printing_type_q(self):
        offset_printing = 0
        if self.is_print:
            q_size = int(self.printer.printing_config['printOffset']) - 1
            # Since when using gpio and not line selection the pulse selects the next print
            offset_printing = 0 if self.printer.printing_config['enableLineSelection'].lower() == 'yes' else 1
        else:
            q_size = 0  # No Printing is needed
        self.printing_type_q = Queue(maxsize=q_size)
        if q_size == 0:
            return

        for k in range(q_size - offset_printing):
            self.printing_type_q.put({'adva': '', 'status': False}, block=False)

    def run(self):
        state = ValidationStates.IDLE
        next_tag_status = False
        self.printer.reset_printer_ack()
        start_test_time = time.time()
        n_skip_location = None
        self.printer_cart_is_ready = False

        while True:
            time.sleep(0)
            try:
                state = self.update_state(state)
                self.logger.info(f'PrintingAndValidation: start state {state}')
                if state == ValidationStates.STOP:
                    self.logger.info('Stop running PrintingAndValidation main loop')
                    break

                elif state == ValidationStates.WAIT_MOH:
                    # wait if Manual Operation Handling is needed:
                    self.wait_for_moh()

                elif state == ValidationStates.WAIT_START:
                    # wait till event start_testing is set
                    self.wait_for_new_event(event=self.test_start_event,
                                            time_to_wait=PrintingAndValidationDefaults.EVENT_WAIT_TIME)

                elif state == ValidationStates.START_TEST:
                    start_test_time = time.time()

                    self.tag_run_location += 1
                    self.logger.info(f'{ASTRIX} PrintingAndValidation: START_TEST: '
                                     f'new tag location {self.tag_run_location} {ASTRIX}')
                    # Stop R2R if long test
                    if self.is_r2r:
                        if self.test.get_total_test_time() > PrintingAndValidationDefaults.MAX_R2R_WAIT_TIME:
                            self.r2r.send_stop_to_r2r()

                elif state == ValidationStates.PRE_VALIDATION:
                    # Validation
                    self.tag_result_to_validate = None
                    if self.validation_q.full() or self.stop_app_event.is_set():
                        n_skip_location, do_skip = self.check_test_end_before_start_validation(n_skip_location)
                        if do_skip:
                            self.logger.info('PrintingAndValidation: PRE_VALIDATION: skip on validation')
                        else:
                            self.tag_result_to_validate = self.validation_q.get(block=False)
                            self.handle_duplication_before_validation()
                            self.add_external_id_to_tag_results()
                    else:
                        self.logger.info('PrintingAndValidation: PRE_VALIDATION: '
                                         'Waiting for the first tag to arrive')
                        if self.is_print and not self.is_scanner and not self.printer.printing_during_movement:
                            is_ack = self.printer.get_printing_ack()
                            if not is_ack:
                                self.logger.warning('pre-validation: printing-only: '
                                                    'did not get printing acknowledgement (PRC), continue anyway')

                elif state == ValidationStates.VALIDATION:
                    if self.tag_result_to_validate is not None:
                        self.validation(self.tag_result_to_validate)

                elif state == ValidationStates.PRINTER_ACK:
                    if self.need_to_check_printer_ack():
                        is_ack = self.printer.get_printing_ack()
                        if not is_ack:
                            raise PrinterException('did not get printing acknowledgement (PRC)')

                elif state == ValidationStates.SET_NEXT_PRINT:  # if bi-directional happens only after test end
                    # select next printing type
                    if self.is_print:
                        if self.printing_type_q.empty():
                            self.logger.info(
                                'PrintingAndValidation: SET_NEXT_PRINT: set next printing to False since queue empty')
                            next_tag_status = False
                            self.logger.info('PrintingAndValidation: SET_NEXT_PRINT: '
                                             'set printing type to false since queue is empty')
                        else:
                            next_tag_to_print = self.printing_type_q.get(block=False)
                            next_tag_status = self.handle_duplication_before_printing(next_tag_to_print)
                        if self.is_print and not self.printer.bi_directional:
                            self.set_next_printing_type(next_tag_status)

                elif state == ValidationStates.WAIT_END:
                    # wait till event end_of_testing is set
                    self.wait_for_new_event(
                        event=self.test_end_event,
                        time_to_wait=self.test.get_total_test_time() + PrintingAndValidationDefaults.EVENT_WAIT_TIME)

                elif state == ValidationStates.END_TEST:
                    # update class queue:
                    if self.tag_results_q.empty():
                        if self.is_print:
                            self.logger.info(f'Empty tag_results_q during {state}')
                            self.add_end_of_test_tags()

                    else:
                        cur_tag_to_print = self.tag_results_q.get(timeout=0.1, block=False)
                        cur_tag_to_print.set_test_info({'tag_run_location': self.tag_run_location,
                                                        'tag_reel_location': self.get_tag_reel_location()})
                        self.validation_q.put(cur_tag_to_print, block=False)
                        if self.is_print:
                            self.printing_type_q.put({'adva': cur_tag_to_print.check_and_get_selected_tag_id(),
                                                      'status': cur_tag_to_print.get_total_test_status()}, block=False)
                    if self.is_print and self.printer.bi_directional:
                        self.set_next_printing_type(next_tag_status)

                elif state == ValidationStates.PRINTER_CART_READY:
                    self.check_if_printer_cart_ready(start_test_time=start_test_time)

                elif state == ValidationStates.NEXT_TAG:
                    # send signal to the r2r to move
                    self.printer.reset_printer_ack()
                    self.printer_cart_is_ready = False
                    if self.is_r2r:
                        self.r2r.move_r2r(is_pass=next_tag_status)
                    else:
                        self.validation_end_event.set()

                elif state == ValidationStates.LOG_DATA:
                    # log data
                    if self.tag_result_to_validate is not None:
                        self.log_data(self.tag_result_to_validate)
                        self.tag_result_to_validate = None
                    if self.is_print:
                        self.update_printer_configs()

                self.logger.info(f'PrintingAndValidation: end state {state}')
            except Exception as e:
                if isinstance(e, PrintingAndValidationException):
                    self.exception_q.put(f'run: state{state.name}: {e}', block=False)
                else:
                    self.exception_q.put(PrintingAndValidationException(f'run: state{state.name}: {e}').__str__(),
                                         block=False)
                if state == ValidationStates.VALIDATION:
                    tag_info = self.tag_result_to_validate.get_test_info()
                    if 'validated' not in tag_info or ValidatedBin[tag_info['validated']].value % 2 == 0:
                        self.tag_result_to_validate.set_test_info({'validated': ValidatedBin.ERROR.name})
                        self.tag_result_to_validate.set_total_fail_bin(fail_code=FailureCodes.SOFTWARE_GENERAL_ERROR,
                                                                       overwrite=True)
                        self.logger.warning(f'PrintingAndValidation: run: state{state.name}: '
                                            f'Error during validation: {tag_info}')
                    else:
                        self.tag_result_to_validate.set_total_fail_bin(fail_code=FailureCodes.BAD_PRINTING)

        self.stop()

    def update_state(self, state):
        # after MOH was done, continue to the next state
        if state == ValidationStates.WAIT_MOH:
            state = self.last_state

        # when event occurred
        if self.stop_app_event.is_set():
            if self.validation_q.empty() and self.tag_results_q.empty():
                if state == ValidationStates.LOG_DATA:
                    state = ValidationStates.STOP
                if state == ValidationStates.END_TEST:
                    state = ValidationStates.NEXT_TAG  # skip NEXT_TAG stage since this is the last iteration
        elif self.moh.get_manual_operation_is_needed():
            self.last_state = state
            state = ValidationStates.WAIT_MOH

        # main flow
        if state == ValidationStates.IDLE:
            state = ValidationStates.WAIT_START
        elif state == ValidationStates.WAIT_START:
            state = ValidationStates.START_TEST
        elif state == ValidationStates.START_TEST:
            state = ValidationStates.PRE_VALIDATION
        elif state == ValidationStates.PRE_VALIDATION:
            state = ValidationStates.VALIDATION
        elif state == ValidationStates.VALIDATION:
            state = ValidationStates.PRINTER_ACK
        elif state == ValidationStates.PRINTER_ACK:
            state = ValidationStates.SET_NEXT_PRINT
        elif state == ValidationStates.SET_NEXT_PRINT:
            state = ValidationStates.WAIT_END
        elif state == ValidationStates.WAIT_END:
            state = ValidationStates.END_TEST
        elif state == ValidationStates.END_TEST:
            state = ValidationStates.PRINTER_CART_READY
        elif state == ValidationStates.PRINTER_CART_READY:
            state = ValidationStates.NEXT_TAG
        elif state == ValidationStates.NEXT_TAG:
            state = ValidationStates.LOG_DATA
        elif state == ValidationStates.LOG_DATA:
            state = ValidationStates.WAIT_START
        return state

    def restart_printer_after_moh(self):
        self.printer.init_printer_hw_connection()
        if self.printer.bi_directional:
            self.init_printer_cart()
    
    def init_printer_cart(self):
        if not self.printer.bi_directional:
            return
        self.printer_cart_is_ready = False
        t_start = time.time()
        dt = 0
        while not self.printer_cart_is_ready and dt < PrintingAndValidationDefaults.PRINTER_90_DEG_WAIT_TIME:
            try:
                self.check_if_printer_cart_ready(t_start)
            except Exception as e:
                self.logger.info(f'init_printer_cart: check_if_printer_cart_ready failed due to {e}, try again')
            self.printer_cart_is_ready = self.printer.current_direction == 'forward'
            dt = time.time() - t_start
        if self.printer.current_direction == 'backward':
            raise PrinterException('restart_printer_after_moh: printer cart is set to the backward direction instead of forward')
        self.printer_cart_is_ready = True
        self.printer.set_printer_direction('forward')
        self.logger.info(f'init_printer_cart: completed successfully')

    def check_if_printer_cart_ready(self, start_test_time):
        if not self.is_print:
            return

        if self.printer.printing_during_movement:
            return

        if self.printer.bi_directional:
            if self.printer_cart_is_ready:
                return
            
            if self.get_gw_signal is None:
                raise PrinterException('check_if_printer_cart_ready: printer is bi-directional but \
                                       get_gw_signal function is not set ')
            gw_signals = self.get_gw_signal()

            if len(gw_signals) == 0:
                raise PrinterException('check_if_printer_cart_ready: no signal was received from gateway')
                
            if PrintingAndValidationDefaults.PRINTER_FORWARD_MSG in gw_signals[-1].lower():
                current_gw_signal = 'forward'
            elif PrintingAndValidationDefaults.PRINTER_BACKWARD_MSG in gw_signals[-1].lower():
                current_gw_signal = 'backward'
            else:
                raise PrinterException(f'check_if_printer_cart_ready: invalid gw signal: {gw_signals[-1]}')
            
            if self.printer.current_direction is not None and self.printer.current_direction == current_gw_signal:
                raise PrinterException(f'check_if_printer_cart_ready: cart signal was not toggled in between '
                                       f'locations. still {current_gw_signal}')
            self.printer.set_printer_direction(current_gw_signal)
            self.printer_cart_is_ready = True
            return
        dt = time.time() - start_test_time
        if dt < PrintingAndValidationDefaults.PRINTER_90_DEG_WAIT_TIME:
            time.sleep(PrintingAndValidationDefaults.PRINTER_90_DEG_WAIT_TIME - dt)

    def check_test_end_before_start_validation(self, n_skip_location):
        if self.stop_app_event.is_set():
            if self.validation_q.full():
                n_skip_location = 0
            elif n_skip_location is None:
                n_skip_location = self.validation_q.maxsize - self.validation_q.qsize()

            if n_skip_location > 0:
                n_skip_location -= 1
                return n_skip_location, True
        return n_skip_location, False

    def handle_duplication_before_printing(self, next_tag):
        is_tag_passed = False
        if not next_tag['status']:
            return is_tag_passed

        duplication_list = self.get_duplicated_tags()
        if next_tag['adva'] != '' and next_tag['adva'] in duplication_list:
            self.change_tag_to_duplicated.append(next_tag['adva'])
            self.logger.info(f'changing printing tag {next_tag["adva"]} to fail due to duplication')
            return is_tag_passed

        is_tag_passed = True
        return is_tag_passed

    def handle_duplication_before_validation(self):
        selected_tag = self.tag_result_to_validate.check_and_get_selected_tag_id()
        if selected_tag in self.change_tag_to_duplicated:
            self.tag_result_to_validate.set_total_fail_bin(FailureCodes.DUPLICATION_OFFLINE)
            self.logger.info(f'changing tag {selected_tag} to duplication fail bin')

    def need_to_check_printer_ack(self):
        return self.is_print and self.is_scanner and self.printer.printing_config['enablePrinterAck'].lower() == 'yes'

    def add_external_id_to_tag_results(self):
        expected_external_id = ''
        configs = self.data_logging.get_log_config()
        if self.is_print:
            if self.tag_result_to_validate.is_all_tests_passed():
                expected_external_id = self.printer.get_expected_external_id(counter=self.external_id_counter)
                self.external_id_counter += 1
            else:
                expected_external_id = ''
                if configs['printingFormat'] == 'prePrint':
                    self.external_id_counter += 1
            self.tag_result_to_validate.set_test_info({'external_id': expected_external_id})
            self.logger.info(f'PrintingAndValidation: assigned external id {expected_external_id}')
        return expected_external_id

    def get_validation_q_size(self):
        return self.validation_q.qsize()

    def get_validation_q_maxsize(self):
        return self.validation_q.maxsize

    def get_tag_reel_location(self):
        configs = self.data_logging.get_log_config()
        offset = int(configs['tag_reel_location']) if 'tag_reel_location' in configs.keys() else 0
        return self.tag_run_location + offset

    def update_scanning_results_after_exception(self):
        if not self.is_scanner:
            return ''
        test_info = {'external_id': 'unknown'}
        try:
            test_info = self.tag_result_to_validate.get_test_info()
            self.validation(self.tag_result_to_validate)
            if self.tag_result_to_validate.get_total_fail_bin() == FailureCodes.BAD_PRINTING.value:
                self.tag_result_to_validate.set_total_fail_bin(fail_code=FailureCodes.PASS, overwrite=True)
        except Exception as e:
            self.logger.warning(f'update_scanning_results_after_exception: {e}')
            return f"Re-Validation for external id {test_info.get('external_id', 'unknown')} Failed: {e}"
        success_str = f"Re-Validation for external id {test_info.get('external_id', 'unknown')} was successful"
        self.logger.info(success_str)
        return success_str

    def set_next_printing_type(self, is_pass):
        """

        @param is_pass:
        @type is_pass: bool or None
        @return:
        @rtype:
        """
        if is_pass is None:
            is_pass = False
        if self.is_print:
            if self.printer.enable_line_selection:
                try:
                    self.printer.set_printing_type(is_tag_passed=is_pass)
                except Exception as e:
                    raise PrinterException(f'set_next_printing_type: {e}')

    def wait_for_new_event(self, event, time_to_wait):
        t_start = datetime.datetime.now()
        while not event.is_set():
            if not self.is_r2r and self.stop_app_event.is_set():
                self.logger.info('stop application during waiting for event while r2r is controlled by gateway')
            elif self.stop_app_event.is_set():
                self.logger.info('stop application during waiting for event')
                time.sleep(PrintingAndValidationDefaults.MIN_WAIT_TIME)
                break
            elif self.moh.get_manual_operation_is_needed():
                self.wait_for_moh()
                t_start = datetime.datetime.now()
                self.first_cycle_after_moh = True
            event.wait(0.1)
            dt = (datetime.datetime.now() - t_start).total_seconds()

            if dt > time_to_wait and not self.moh.get_manual_operation_is_needed():
                raise PrintingAndValidationException('wait_for_new_event: did not receive ack for '
                                                     'starting/ending tag test')
        event.clear()

    def wait_for_moh(self):
        try:
            if self.is_r2r:
                self.r2r.send_stop_to_r2r()
        except Exception as e:
            self.logger.warning(f'Could not stop R2R during wait for manual operation handling due to {e}')
        self.logger.info('PrintingAndValidation: wait for manual operation handling')
        while True:
            time.sleep(1)
            if not self.moh.get_manual_operation_is_needed():
                break
            if self.stop_app_event.is_set():
                break

    def get_estimated_external_id(self):
        validation_loc = self.tag_run_location - self.get_validation_q_maxsize()
        if validation_loc >= len(self.scanning_data):
            raise Exception('tag location does not exist in the specified file for scanning without printing')
        expected_value = self.scanning_data.iloc[validation_loc]['external_id']
        expected_value = expected_value if not pd.isnull(expected_value) else ''
        return expected_value

    def validation(self, expected_result):
        """

        @param expected_result:
        @type expected_result: WiliotTesterTagResultList
        @return:
        @rtype:
        """
        is_validated = ValidatedBin.UNTESTED
        test_info = {"external_id": ''}
        hw_exc = ''

        if self.is_print and self.is_scanner:
            # scan and compare
            test_info = expected_result.get_test_info()
            try:
                is_validated, res_str = self.scanner.validated_scan_data(expected_value=test_info['external_id'])
                hw_exc = f' [{res_str}]'
            except Exception as e:
                is_validated = ValidatedBin.NO_READ
                hw_exc = f' [{e}]'

        elif self.is_print and not self.is_scanner:
            # get_status from printer
            if not self.printer.printing_during_movement:
                is_ack = self.printer.get_printing_ack()
                if not is_ack:
                    self.logger.warning('printing only - did not get printing acknowledgement (PRC), '
                                        'try to get status anyway')
            status = self.printer.get_printer_status(counter=self.get_r2r_counter(),
                                                     is_pass=expected_result.is_all_tests_passed())
            if status:
                is_validated = ValidatedBin.PRINT_ACK_ONLY
            else:
                is_validated = ValidatedBin.MISMATCH

        elif not self.is_print and self.is_scanner:
            try:
                expected_value = self.get_estimated_external_id()
                expected_result.set_test_info({'external_id': expected_value})
                test_info["external_id"] = expected_value
                is_validated, res_str = self.scanner.validated_scan_data(expected_value=expected_value)
                hw_exc = f' [{res_str}]'
            except Exception as e:
                is_validated = ValidatedBin.NO_READ
                hw_exc = f' [{e}]'

        else:  # no print and no scanner
            pass

        expected_result.set_test_info({'validated': is_validated.name})
        if is_validated.value % 2 == 1:
            id_str = f'id: {test_info["external_id"]}' if test_info["external_id"] is not None else 'black square'
            raise PrintingAndValidationException(
                f'Validation Failed due to: {is_validated} for {id_str}\n{hw_exc}')

    def log_data(self, new_tag_results):
        # update packet data
        self.data_logging.update_packets_data(res=new_tag_results)
        # update run data
        self.data_logging.update_run_data(res=new_tag_results)

    def add_end_of_test_tags(self):
        if self.is_print:
            dummy_dict = {'tag_run_location': self.tag_run_location,
                          'tag_reel_location': self.get_tag_reel_location(),
                          'fail_bin': FailureCodes.END_OF_TEST.value,
                          'fail_bin_str': FailureCodes.END_OF_TEST.name,
                          'trigger_time': None, 
                          'total_test_duration': None,
                          'crc_environment_previous': None,
                          'label_validated': 'UNTESTED'}
            self.end_of_test_tags.append(dummy_dict)

    def add_end_of_test_to_log(self):
        for dummy_tag in self.end_of_test_tags:
            # Update only packet data
            self.data_logging.update_default_test_data(dummy_tag)
            self.data_logging.update_packets_data()

    def update_printer_configs(self):
        # update printing config:
        configs = self.data_logging.get_log_config()
        if not configs['toPrint'].lower() == 'yes':
            return
        file_name = get_print_user_config_file(configs['printingFormat'])
        config_path = os.path.join(self.test.dir_config, file_name)
        data = open_json_cache(folder_path=self.test.dir_config,
                               file_path=config_path,
                               default_values=DefaultGUIValues(configs['printingFormat']).default_gui_values)
        data['firstPrintingValue'] = self.printer.get_expected_external_id(
            counter=self.external_id_counter)[-PrintingAndValidationDefaults.TAG_COUNT_SIZE:]
        data['tagLocation'] = str(self.tag_run_location)
        data['tag_reel_location'] = str(self.get_tag_reel_location() +
                                        PrintingAndValidationDefaults.R2R_START_OFFSET)
        with open(config_path, "w") as f:
            json.dump(data, f)

    def stop(self):
        if self.end_of_test_tags:
            self.add_end_of_test_to_log()
        if self.printer.enable:
            self.update_printer_configs()
            self.printer.exit()
        if self.scanner.enable:
            self.scanner.exit()
        if self.is_r2r:
            self.r2r.exit()


class Printer(object):
    def __init__(self, is_print, exception_queue, logger_name, printing_config=None):
        """

        @param exception_queue:
        @type exception_queue: Queue
        @param logger_name:
        @type logger_name: str
        @param printing_config:
        @type printing_config: dict
        """
        self.exception_queue = exception_queue
        self.logger = logging.getLogger(logger_name)
        self.printing_config = printing_config
        self.got_printer_ack = False
        self.enable = True
        if printing_config is None or not is_print:
            self.logger.info('Printer is disable')
            self.enable = False
            return

        self.printer_socket = None
        self.dir_config = os.path.abspath(os.path.join(os.path.dirname(__file__), '../configs'))

        self.enable_line_selection = self.printing_config['enableLineSelection'].lower() == 'yes'
        self.printing_during_movement = self.printing_config['printingDuringMovement'].lower() == 'yes'
        self.bi_directional = self.printing_config['printingBidirectionalMovement'].lower() == 'yes'
        self.printed_id = 0
        self.current_direction = None
        self.cur_value = 0

        try:
            printer_configs = self.set_printer_config()
            self.printing_config = {**self.printing_config, **printer_configs}
            self.init_printer_hw_connection()
            self.init_printing_protocols()
            self.logger.info('PRINTER: printer is ready after initialization')
        except Exception as e:
            raise PrinterException(f'init: {e}')

    def init_printer_hw_connection(self):
        self.open_printer_socket()
        self.logger.info('PRINTER: printer is connected')
        self.set_printer_to_running()
        self.logger.info('PRINTER: printer is running')
        self.current_direction = None

    def get_expected_external_id(self, counter=0):
        if self.enable:
            tag_counter = counter + int(self.printing_config['firstPrintingValue']
                                        if 'firstPrintingValue' in self.printing_config.keys() else 0)
            printed_external_id, is_ok = get_printed_value(
                self.printing_config['stringBeforeCounter'],
                str(tag_counter))
            if not is_ok:
                raise PrinterException('get_expected_external_id: failed to get a valid printed value')
            return printed_external_id

    def set_printer_direction(self, direction):
        self.current_direction = direction

    def set_printer_config(self):
        try:
            defaults = ConfigDefaults()

            file_config = os.path.abspath(os.path.join(self.dir_config, 'configs_for_printer_values.json'))
            printer_configs = open_json_cache(self.dir_config, file_config, defaults.get_printer_defaults())
        except Exception as e:
            raise PrinterException(f'init: set_printer_config: {e}')
        return printer_configs

    def is_connected(self):
        try:
            if self.printer_socket is None:
                return False
            self.query(self.get_state_request())
            return True
        except Exception as e:
            return False

    def open_printer_socket(self):
        """
        opens the printer socket
        """
        try:
            if not self.is_connected():
                self.printer_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.printer_socket.settimeout(PrintingAndValidationDefaults.PRINTER_SOCKET_TIMEOUT)
                self.printer_socket.connect((self.printing_config['TCP_IP'],
                                             int(self.printing_config['TCP_PORT'])))
        except Exception as e:
            raise PrinterException(f'could not open_printer_socket: {e}')

    def exit(self):
        try:
            time.sleep(4)  # wait for printer to finish last print
            self.query(self.set_state_command('2'))  # for regular closure (not when connection error happens)
            self.printer_socket.close()
        except Exception as e:
            self.logger.warning(f'Could not exit printer due to: {e}')

    def set_printer_to_running(self):
        """
        sets the printer to running mode
        Zipher Text Communications Protocol
        printer state machine:
           0 -> 1                      shutdown
           1 -> 4 (automatically)      starting-up
           2 -> 0 (automatically)      shutting-down
           3 -> 2, 4                   running
           4 -> 2, 3                   offline
        @except: PrinterNeedsResetException('Printer failed to switch to running mode')
        @return: None
        """
        res = self.query(self.get_state_request())
        parts = res.split('|')
        if len(parts) < 2:
            raise PrinterException(f'set_printer_to_running: got unexpected results {res}')
        if parts[1] == '0':  # (Shut down)
            res = self.query(self.set_state_command('1'))
            if res == 'ACK':
                counter = 0
                while counter < 10:
                    time.sleep(1)
                    res = self.query(self.set_state_command('3'))
                    if res == 'ACK':
                        return
                    counter += 1
        elif parts[1] == '3':  # (Running)
            return
        elif parts[1] == '4':  # (Offline)
            res = self.query(self.set_state_command('3'))
            if res == 'ACK':
                return
        e = 'Printer failed to switch to running mode'
        raise PrinterException(f'set_printer_to_running: {e}')

    def query(self, cmd):
        """Send the input cmd string via TCPIP Socket
        @type cmd: str
        @param cmd: command to send to printer
        @return: the reply string
        """
        cmd_str = cmd.strip('\r\n')
        self.logger.info(f'Sent command to printer: {cmd_str}')
        self.printer_socket.send(cmd.encode())
        values = self.read(cmd=cmd)
        if 'PRC' in values:
            self.got_printer_ack = True
            if len(set(values)) == 1:  # only PRC was received
                values = self.read(cmd=cmd)
        for value in values:
            if value != 'PRC' and value != '':
                return value

        raise Exception(f'Printer: query: did not ack to the following cmd :{cmd}')

    def read(self, cmd=''):
        data = self.printer_socket.recv(int(self.printing_config['TCP_BUFFER']))
        self.logger.info(f'Received RAW answer from printer: {data}')
        if len(data) == 0:
            raise PrinterException(f'query: did not ack to the following cmd :{cmd}' if cmd else '')
        value = data.decode("utf-8")
        str_split = '\r\n'
        if str_split not in value:
            str_split = '\r' if '\r' in value else '\n'

        values = value.split(str_split)[:-1]
        self.logger.info(f'Received answer from printer: {values}')
        return values

    def reset_printer_ack(self):
        self.got_printer_ack = False

    def get_printing_ack(self):
        if self.got_printer_ack:
            self.got_printer_ack = False
            return True
        t_i = datetime.datetime.now()
        dt = 0.0
        while dt < PrintingAndValidationDefaults.PRINTER_ACK_TIMEOUT:
            try:
                values = self.read()
                if 'PRC' in values:
                    return True
                raise PrinterException(f'while waiting for ack, received a different response: {values}')
            except Exception as e:
                self.logger.info(f'Printer got exception, try again ({e})')
                dt = (datetime.datetime.now() - t_i).total_seconds()
                time.sleep(PrintingAndValidationDefaults.TIME_BTWN_PRINTER_REQUESTS)
        return False

    def get_printer_status(self, counter, is_pass):
        t_i = datetime.datetime.now()
        dt = 0.0
        while dt < PrintingAndValidationDefaults.PRINTER_ACK_TIMEOUT:
            try:
                status = self.printer_status(counter=self.cur_value + counter, is_pass=is_pass)
                if status:
                    return True
                dt = (datetime.datetime.now() - t_i).total_seconds()
                time.sleep(PrintingAndValidationDefaults.TIME_BTWN_PRINTER_REQUESTS)
            except Exception as e:
                raise PrinterException(f'get_printer_status: exception:({e})')

        return False

    @staticmethod
    def get_state_request():
        """
        gets the situation of the printer
        @return: the situation in the printer in the following format:
            STS|<overallstate>|<errorstate>|<currentjob>|<batchcount>|<totalcount>|<
        """
        cmd = 'GST\r\n'
        return cmd

    @staticmethod
    def set_state_command(desired_state):
        """
        builds the command to send to printer for setting a printer state
        @param desired_state: the state to enter to, according to the following description
        0 Shut down
        1 Starting up
        2 Shutting down
        3 Running
        4 Offline
        @return: the cmd to send to printer
        """
        cmd = 'SST|' + str(desired_state) + '|\r\n'
        return cmd

    def get_line_assignment_cmd(self, line, tag_counter):
        job_name = self.printing_config['passJobName']
        if 'SGTIN' in job_name.upper():
            raise Exception('biderectional mirror printing is not supported for SGTIN')
        job_name += '_M' if line == PrintingAndValidationDefaults.PASS_MIRROR_JOB_NUM else ''
        reel_id = self.printing_config["stringBeforeCounter"]
        return f'LAS|{job_name}|{line}|reel_num={reel_id}T|tag_number={str(tag_counter).zfill(4)}|\r\n'

    def get_line_number(self, is_pass):
        if is_pass:
            if self.current_direction is None or self.current_direction == 'forward':
                line = PrintingAndValidationDefaults.PASS_JOB_NUM
            elif self.current_direction == 'backward':
                line = PrintingAndValidationDefaults.PASS_MIRROR_JOB_NUM
            else:
                raise Exception(f'invalid printer direction: {self.current_direction}')
        else:
            line = PrintingAndValidationDefaults.FAIL_JOB_NUM
        return line
    
    @staticmethod
    def get_line_selection_cmd(line):
        return f"LSL|{line}|\r\n"

    def set_job_fields(self, tag_num_offset=0):
        pass_job_name = self.printing_config['passJobName']
        fail_job_name = self.printing_config['failJobName']
        reel_id = self.printing_config["stringBeforeCounter"]

        tag_num_int = int(self.printing_config["firstPrintingValue"]) + tag_num_offset
        tag_num_str = str(tag_num_int).zfill(4)
        if 'BARCODE' in pass_job_name.upper():
            job_fields = [f'reel_num={reel_id}T']
        elif 'SGTIN' in pass_job_name.upper():
            job_fields = [f'sgtin={reel_id[:18]}',
                          f'reel_num={reel_id[18:26]}T']
        elif 'BLANK' in pass_job_name.upper():
            job_fields = []
        else:
            raise PrinterException('Job Name inserted is not supported at the moment')

        if pass_job_name != 'BLANK':
            job_fields.append(f'tag_number={tag_num_str}')
        job_fields.append('\r\n')
        job_fields_str = '|'.join(job_fields)

        pass_job_fields_cmd = f'LAS|{pass_job_name}|{PrintingAndValidationDefaults.PASS_JOB_NUM}|' + job_fields_str
        fail_job_fields_str = '\r\n' if fail_job_name != pass_job_name else job_fields_str
        fail_job_fields_cmd = f'LAS|{fail_job_name}|{PrintingAndValidationDefaults.FAIL_JOB_NUM}|' + fail_job_fields_str
        all_cmds = [pass_job_fields_cmd, fail_job_fields_cmd]
        if self.bi_directional:
            all_cmds.append(
                f'LAS|{pass_job_name}_M|{PrintingAndValidationDefaults.PASS_MIRROR_JOB_NUM}|' + job_fields_str)

        return all_cmds

    def init_printing_protocols(self, tag_num_offset=0):

        try:
            job_fields_cmds = self.set_job_fields(tag_num_offset=tag_num_offset)

            cmds = ['CAF\r\n',
                    'CQI\r\n',
                    f'CLN|{PrintingAndValidationDefaults.PASS_JOB_NUM}|\r\n',
                    f'CLN|{PrintingAndValidationDefaults.FAIL_JOB_NUM}|\r\n',
                    ]
            if self.bi_directional:
                cmds += [f'CLN|{PrintingAndValidationDefaults.PASS_MIRROR_JOB_NUM}|\r\n']
            cmds += job_fields_cmds

            for cmd in cmds:
                value = self.query(cmd)
                time.sleep(0.1)
                # check if the return value is good, if not retry again for 10 times
                counter = 0
                while counter < 10:
                    # 'CQI' fails if the queue is empty
                    if value == 'ERR' and 'CQI' not in cmd:
                        counter += 1
                        time.sleep(0.1)
                        value = self.query(cmd)
                    else:
                        break
                if counter >= 10:
                    raise PrinterException('initialization process has failed in command: ' + cmd)
            # get the current counter value
            value = self.query(self.get_state_request())
            if value == 'ERR':
                raise PrinterException('initialization process has failed in command: ' + self.get_state_request())
            else:
                parts = [p for p in value.split("|")]
                if len(parts) < 6:
                    raise PrinterException(f'init_printing_protocols: got unexpected results {value}')
                self.cur_value = int(parts[5])

        except Exception as e:
            raise PrinterException(f'init_printing_protocols: {e}')

    def set_line_selection(self, line_number):
        """
        line selection
        @return:
        @rtype:
        """
        try:
            line_selection_msg = self.get_line_selection_cmd(line_number)
            rsp = ''
            t_0 = datetime.datetime.now()
            dt = 0
            timeout_ack = 1
            while rsp != 'ACK' and dt < timeout_ack:
                rsp = self.query(line_selection_msg)
                if rsp == 'ACK':
                    return
                dt = (datetime.datetime.now() - t_0).total_seconds()
            raise Exception('getting ERR from printer')
        except Exception as e:
            raise PrinterException(f'set_line_selection: Did not get ACK for Line Selection: {e}')
    
    def set_update_line_data(self, line):
        """
        line selection
        @return:
        @rtype:
        """
        try:
            update_line_msg = self.get_line_assignment_cmd(line=line, tag_counter=int(self.printing_config["firstPrintingValue"]) + self.printed_id)
            rsp = ''
            t_0 = datetime.datetime.now()
            dt = 0
            timeout_ack = 1
            while dt < timeout_ack:
                rsp = self.query(update_line_msg)
                if rsp and rsp != 'ERR':
                    return
                dt = (datetime.datetime.now() - t_0).total_seconds()
            raise Exception('getting ERR from printer')
        except Exception as e:
            raise PrinterException(f'set_update_line_data: Did not get ACK for Update Line Data: {e}')
    
    def set_printing_type(self, is_tag_passed):
        """
        line selection
        @return:
        @rtype:
        """
        line_number = self.get_line_number(is_tag_passed)
        if self.bi_directional and is_tag_passed:
            self.set_update_line_data(line_number)
        self.set_line_selection(line_number)
        self.printed_id += int(is_tag_passed)

    def printer_status(self, counter, is_pass):
        """
        checks if the printing value matches the values registered to the logs
        should be called only after self.events.r2r_ready was set
        Exceptions:
            @except Exception('The printer printed Pass to the previous tag'):
                    printer printed pass while it should have been print fail
            @except Exception('The printer printed Fail to the previous tag')
                    printer printed fail while it should have been print pass
            @except Exception('The printer have not printed on the last tag')
                    printer did not print while it should have been
        """
        res = self.query(
            self.get_state_request())  # STS|<overallstate>|<errorstate>|<currentjob>|<batchcount>|<totalcount>
        parts = [p for p in res.split("|")]
        if len(parts) < 6:
            raise PrinterException(f'printer_status: got unexpected results {res}')
        if parts[1] == '3':
            if parts[2] == '0':
                if int(parts[5]) == counter:
                    if not self.printing_config['printingFormat'] != 'Test':
                        job_name = self.printing_config['passJobName'] if is_pass else \
                            self.printing_config['failJobName']

                        if job_name in parts[3]:
                            self.logger.debug(f'Compare success with type {job_name} and counter {counter}')
                            return True
                        else:
                            self.logger.info(f'Printer type was not right (expected: {job_name}, got: {parts[3]}')
                            return False
                    else:
                        self.logger.debug(f'Compare success with counter {counter}')
                        return True
                else:
                    self.logger.info(f'The printer counter is not synced (expected: {counter}, got: {parts[5]}'
                                     f', trying again')
                    return False
            else:
                if parts[2] == '1':
                    raise PrinterException('error-state is Warnings present')
                if parts[2] == '2':
                    raise PrinterException('error-state is Faults present')
        else:
            if parts[1] == '0':
                raise PrinterException('over-all-state is Shutdown')
            if parts[1] == '1':
                raise PrinterException('over-all-state is Starting up')
            if parts[1] == '2':
                raise PrinterException('over-all-state is Shutting down')
            if parts[1] == '4':
                raise PrinterException('over-all-state is Offline')
            if parts[2] == '1':
                raise PrinterException('error-state is Warnings present')
            if parts[2] == '2':
                raise PrinterException('error-state is Faults present')


class Scanner(object):
    def __init__(self, is_scanner, exception_queue, logger_name, scanning_config, is_print=False):
        """

        @param exception_queue:
        @type exception_queue:
        @param logger_name:
        @type logger_name:
        @param scanning_config: all scanner configuration
        @type scanning_config: dict
        @param is_print: True if run with printing
        @type is_print: bool
        """
        self.exception_queue = exception_queue
        self.logger = logging.getLogger(logger_name)

        self.enable = True
        if scanning_config is None or not is_scanner:
            self.logger.info('Scanner is disable')
            self.enable = False
            return

        self.scanning_config = scanning_config

        scanner_port = f'COM{scanning_config["QRcomport"]}'
        try:
            if self.scanning_config["scannerType"].lower() == 'rtscan':
                self.scanner = BarcodeScanner(com_port=scanner_port, log_type='LOG', write_to_log=not is_print,
                                              timeout=str(self.scanning_config['QRtimeout']),
                                              is_flash=self.scanning_config['scannerFlash'])
            elif self.scanning_config["scannerType"].lower() == 'cognex':
                self.scanner = CognexDataMan(port=scanner_port, timeout=self.scanning_config['QRtimeout'])
            else:
                raise Exception('scanner type is not supported. Please use rtscan or cognex')
            self.logger.info('Connected to scanner at port {}'.format(scanner_port))
        except Exception as e:
            raise ScannerException(f'init: could not connect to barcode scanner {e}')

    def reconnect(self):
        try:
            if not self.scanner.is_open():
                self.scanner.open_port(com_port=self.scanner.com_port, timeout=str(self.scanning_config['QRtimeout']))
        except Exception as e:
            raise ScannerException(f'reconnect: could not connect to barcode scanner {e}')

    def validated_scan_data(self, expected_value=None):
        validated_status = ValidatedBin.UNTESTED
        t_i = datetime.datetime.now()
        dt = 0.0
        is_first = True
        while dt < PrintingAndValidationDefaults.SCANNER_GOOD_READ_TIMEOUT:
            try:
                scanned_data, _, _, _ = \
                    self.scanner.scan_ext_id(
                        need_to_trigger=is_first or PrintingAndValidationDefaults.TRIGGER_TYPE == 0)
                res_str = f'expected:{expected_value}, scanned:{scanned_data}'
                self.logger.info(f'{ASTRIX} SCANNING: {res_str} {ASTRIX}')
                if expected_value is None:
                    validated_status = ValidatedBin.UNTESTED
                elif scanned_data == expected_value or (expected_value == '' and scanned_data is None):
                    return ValidatedBin.MATCH, res_str
                elif scanned_data is None:
                    validated_status = ValidatedBin.NO_READ
                else:
                    validated_status = ValidatedBin.MISMATCH

                dt = (datetime.datetime.now() - t_i).total_seconds()
                time.sleep(PrintingAndValidationDefaults.TIME_BTWN_PRINTER_REQUESTS)
                is_first = False
            except Exception as e:
                self.scanner.trigger_off()
                raise ScannerException(f'get_scan_data: got exception from scanner ({e})')
        self.scanner.trigger_off()
        return validated_status, res_str

    def exit(self):
        try:
            self.scanner.close_port()
        except Exception as e:
            self.logger.warning(f'Could not exit scanner due to: {e}')



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import importlib
import sys

from wiliot_testers.yield_tester.configs.inlay_data import all_inlays
from wiliot_testers.yield_tester.simulation.yield_simulation_utils import get_simulated_gw_port, AUTO_PACKET, \
    AUTO_TRIGGERS, TIME_BETWEEN_AUTO_TRIGGERS
from wiliot_core import *

script_dir = os.path.dirname(__file__)
json_file_path = os.path.join(script_dir, '../configs', 'user_inputs.json')
default_user_inputs = {
    "min_cumulative": "60",
    "min_cumulative_line": "yes",
    "min_current": "20",
    "min_current_line": "yes",
    "max_temperature": "40",
    "min_temperature": "10",
    "temperature_type": "C",
    "min_humidity": "20",
    "max_humidity": "90",
    "min_light_intensity": "0",
    "max_light_intensity": "1500",
    "red_line_cumulative": "85",
    "red_line_current": "50",
    "pin_number": "004",
    "Arduino": "Yes",
    "rssi_threshold": ""
}
try:
    with open(json_file_path) as f:
        user_inputs = json.load(f)
    for key, value in default_user_inputs.items():
        if key not in user_inputs:
            user_inputs[key] = value
    with open(json_file_path, 'w') as f:
        json.dump(user_inputs, f, indent=4)
except Exception as e:
    user_inputs = default_user_inputs
    os.makedirs(os.path.dirname(json_file_path), exist_ok=True)
    with open(json_file_path, 'w') as f:
        json.dump(user_inputs, f, indent=4)

SECONDS_WITHOUT_PACKETS = 60
SECONDS_FOR_GW_ERROR_AFTER_NO_PACKETS = 120
MAX_SUB1G_POWER = 29
MAX_BLE_POWER = 22
ARDUINO_EXISTS = (user_inputs.get('Arduino', '').strip().lower() == 'yes')

inlay_data_eng_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'configs', 'inlay_data_eng.py')
if os.path.exists(inlay_data_eng_path):
    inlay_data = getattr(importlib.import_module('wiliot_testers.yield_tester.configs.inlay_data_eng'), 'all_inlays', {})
    inlays = inlay_data if inlay_data else all_inlays
else:
    inlays = all_inlays


class AdvaProcess(object):
    """
    Counting the number of unique advas
    """

    def __init__(self, stop_event, inlay_type, logging_file, listener_path, time_between_matrices, rssi_th):
        self.stopped_by_user = False
        self.take_care_of_pausing = False
        self.gw_error_connection = False
        self.second_without_packets = False
        self.gw_instance = None
        self.rssi_th = rssi_th
        self.logger_file = logging_file
        self.listener_path = listener_path
        self.all_tags = Queue()
        self.stop = stop_event
        self.gw_start_time = datetime.datetime.now()
        self.init_gw(listener_path)
        self.trigger_time_queue = deque(maxlen=10)
        self.last_change_time = 0
        self.number_of_sensor_triggers = 0
        self.needed_time_between_matrices = TIME_BETWEEN_AUTO_TRIGGERS if AUTO_TRIGGERS else time_between_matrices
        self.inlay_type = inlay_type
        self.gw_reset_config()
        time.sleep(1)

    def init_gw(self, listener_path=None):

        try:
            if self.gw_instance is None:
                gw_port = get_simulated_gw_port() if AUTO_PACKET else None
                self.gw_instance = WiliotGateway(auto_connect=True,
                                                 logger_name='yield',
                                                 log_dir_for_multi_processes=listener_path,
                                                 port=gw_port,
                                                 np_max_packet_in_buffer_before_error=10)

            else:
                # reconnect
                is_connected = self.gw_instance.is_connected()
                if is_connected:
                    self.gw_instance.close_port()
                self.gw_instance.open_port(self.gw_instance.port, self.gw_instance.baud)

            is_connected = self.gw_instance.is_connected()
            if is_connected:
                self.gw_instance.start_continuous_listener()
            else:
                self.logger_file.warning("Couldn't connect to GW in main thread")
                raise Exception(f"Couldn't connect to GW in main thread")

        except Exception as ee:
            raise Exception(f"Couldn't connect to GW in main thread, error: {ee}")

    def set_stopped_by_user(self, stopped):
        self.stopped_by_user = stopped
        self.take_care_of_pausing = True

    def get_gw_start_time(self):
        return self.gw_start_time

    def get_last_change_time(self):
        return self.last_change_time

    def get_gw_error_connection(self):
        return self.gw_error_connection

    def pop_from_triggers_queue(self):
        self.trigger_time_queue.popleft()

    def get_trigger_time_queue(self):
        return self.trigger_time_queue

    def get_sensors_triggers(self):
        return self.number_of_sensor_triggers

    def gw_reset_config(self, start_gw_app=False):
        """
        Configs the gateway
        """
        if self.gw_instance.connected:
            self.gw_instance.reset_gw()
            self.gw_instance.reset_listener()
            time.sleep(2)
            if not self.gw_instance.is_gw_alive():
                self.logger_file.warning('gw_reset_and_config: gw did not respond')
                raise Exception('gw_reset_and_config: gw did not respond after rest')

            gw_config = inlays.get(self.inlay_type)

            cmds = {CommandDetails.scan_ch: gw_config['received_channel'],
                    CommandDetails.time_profile: gw_config['time_profile_val'],
                    CommandDetails.set_energizing_pattern: gw_config['energy_pattern_val'],
                    CommandDetails.set_sub_1_ghz_power: [MAX_SUB1G_POWER],
                    CommandDetails.set_scan_radio: self.gw_instance.get_cmd_symbol_params(
                        freq_str=gw_config['symbol_val']),
                    CommandDetails.set_rssi_th: self.rssi_th,
                    }
            output_power_cmds = self.gw_instance.get_cmds_for_abs_output_power(abs_output_power=MAX_BLE_POWER)
            cmds = {**cmds, **output_power_cmds}
            self.gw_instance.set_configuration(cmds=cmds, start_gw_app=start_gw_app, read_max_time=1)
            if not ARDUINO_EXISTS and not AUTO_TRIGGERS:
                pin_num = user_inputs.get('pin_number')
                cmd = '!cmd_gpio CONTROL_IN P%s 0' % pin_num.zfill(3)
                self.gw_instance.write(cmd, must_get_ack=True)
        else:
            raise Exception('Could NOT connect to GW')

    def raising_trigger_number(self):
        self.number_of_sensor_triggers += 1
        self.logger_file.info(f'Got a Trigger.  Number of Triggers {self.number_of_sensor_triggers}')

    def run(self):
        """
        Receives available data then counts and returns the number of unique advas.
        """
        self.gw_instance.set_configuration(start_gw_app=True)
        self.gw_instance.reset_start_time()
        self.gw_start_time = datetime.datetime.now()
        got_new_adva = False
        no_data_start_time = None  # Time when we first detect no data available

        while not self.stop.is_set():
            time.sleep(0)
            current_time_of_data = (datetime.datetime.now() - self.gw_start_time).total_seconds()
            time_condition_met = current_time_of_data - self.last_change_time >= self.needed_time_between_matrices

            gw_rsp = self.gw_instance.get_gw_rsp()

            if not self.stopped_by_user and self.take_care_of_pausing:
                self.gw_reset_config(start_gw_app=True)
                self.take_care_of_pausing = False
            elif self.stopped_by_user and self.take_care_of_pausing:
                self.gw_instance.reset_gw()
                self.take_care_of_pausing = False

            if time_condition_met:
                if AUTO_TRIGGERS:
                    self.last_change_time = (datetime.datetime.now() - self.gw_start_time).total_seconds()  # a float
                    self.raising_trigger_number()
                    self.trigger_time_queue.append([self.last_change_time, self.number_of_sensor_triggers])
                else:
                    if gw_rsp is not None and 'Detected High-to-Low peak' in gw_rsp['raw'] and not self.stopped_by_user:
                        self.last_change_time = gw_rsp['time']  # a float
                        self.raising_trigger_number()
                        self.trigger_time_queue.append([self.last_change_time, self.number_of_sensor_triggers])

            if self.gw_instance.is_data_available() and not self.stopped_by_user:
                raw_packets_in = self.gw_instance.get_packets(action_type=ActionType.ALL_SAMPLE,
                                                              data_type=DataType.RAW, tag_inlay=self.inlay_type)
                if not self.all_tags.full():
                    self.all_tags.put(raw_packets_in)
                else:
                    self.logger_file.warning(f"Queue is full.. Packet: {raw_packets_in}")
                got_new_adva = True
                no_data_start_time = None
            else:
                if not self.stopped_by_user:
                    if no_data_start_time is None:
                        no_data_start_time = time.time()
                    if time.time() - no_data_start_time >= SECONDS_WITHOUT_PACKETS:
                        got_new_adva = False
                        if not self.second_without_packets:
                            self.logger_file.warning("One minute without packets..")
                            self.second_without_packets = True
                        time.sleep(5)
                        if not self.gw_instance.is_connected():
                            self.reconnect()
                    if time.time() - no_data_start_time >= SECONDS_FOR_GW_ERROR_AFTER_NO_PACKETS:
                        self.gw_error_connection = True
                        break
                    if self.gw_instance.get_read_error_status():
                        self.logger_file.warning("Reading error.. Listener did recovery flow.")
                    time.sleep(0.050 if not got_new_adva else 0)
                else:
                    no_data_start_time = None
        self.gw_instance.reset_gw()
        self.gw_instance.exit_gw_api()

    def reconnect(self):
        self.logger_file.info('Trying to reconnect to GW')
        try:
            self.init_gw()
            self.gw_reset_config(start_gw_app=True)
        except Exception as e:
            self.logger_file.warning(f"Couldn't reconnect GW, due to: {e}")

    def get_raw_packets_queue(self):
        """
        Returns the packet queue that is created above
        """
        return self.all_tags



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import os
import time

import pandas as pd
from wiliot_tools.utils.wiliot_gui.wiliot_gui import WiliotGui


def get_files_path():
    """
    opens GUI for selecting a file and returns it
    """
    layout_dict = {
        'file_path': {'text': 'select file for maunal removal:', 'value': '',
                      'widget_type': 'file_input'}}
    gui = WiliotGui(params_dict=layout_dict)
    values_out = gui.run()
    if not os.path.isfile(values_out['file_path']) or values_out['file_path'] == '':
        raise Exception(f'file: {values_out["file_path"]} does not exist')
    loc_df = pd.read_csv(values_out['file_path'])
    return loc_df


def print_tags_to_remove(msg_to_print):
    ind = 0
    next_prev_layout_dict = {
        'msg': {'value': msg_to_print[ind], 'widget_type': 'label'},
        'next_button': {'text': 'Next', 'value': '', 'widget_type': 'button'},
        'prev_button': {'text': 'Previous', 'value': '', 'widget_type': 'button'}
    }

    def next_func():
        nonlocal ind
        ind += 1
        if ind >= len(msg_to_print):
            print('end of run')
            time.sleep(1)
            gui.layout.quit()
            gui.layout.destroy()
        gui.update_widget(widget_key='msg', new_value=msg_to_print[ind])
        return

    def prev_func():
        nonlocal ind
        ind -= 1 if ind > 0 else 0
        gui.update_widget(widget_key='msg', new_value=msg_to_print[ind])

    gui = WiliotGui(params_dict=next_prev_layout_dict, do_button_config=False)

    gui.add_event(widget_key='next_button', command=next_func, event_type='button')
    gui.add_event(widget_key='prev_button', command=prev_func, event_type='button')
    gui.run()


def extract_user_msg(df):
    """

    @param df:
    @type df: pd.DateFrame
    @return:
    @rtype:
    """
    df.sort_values(by='location', ascending=False, inplace=True)
    df = df[~((df["asset_code"].isna()) & (df["wiliot_code"].isnull()))]
    df.reset_index(inplace=True)
    msg_out = [f'First Label should be: {df["asset_code"].iloc[0]}, {df["wiliot_code"].iloc[0]}']
    last_good_label = ''
    bads_label = []
    n_bad_labels = 0
    for ind_num, row in df.iterrows():
        if row['is_success']:
            if bads_label:
                if len(bads_label) == 1:
                    bads_label_list = f'{bads_label[0]["asset_code"]}, {bads_label[0]["wiliot_code"]}\n'
                else:
                    bads_label_list = f'from: {bads_label[0]["asset_code"]}, {bads_label[0]["wiliot_code"]}\n' \
                                      f'till: {bads_label[-1]["asset_code"]}, {bads_label[-1]["wiliot_code"]}'

                msg_out.append(f'good label BEFORE: {last_good_label}\n\n'
                               f'Please remove the following ({len(bads_label)} tags):\n'
                               f'{bads_label_list}\n\n'
                               f'good label AFTER: {row["asset_code"]}, {row["wiliot_code"]}\n\n')

            last_good_label = f'{row["asset_code"]}, {row["wiliot_code"]}'
            bads_label = []
            continue
        # bad tag
        bads_label.append(row[['asset_code', 'wiliot_code']])
        n_bad_labels += 1

    msg_out.append(f'Last Label should be: {df["asset_code"].iloc[-1]}, {df["wiliot_code"].iloc[-1]}')
    return msg_out, n_bad_labels


if __name__ == '__main__':
    location_df = get_files_path()
    user_msgs, n_bad_labels = extract_user_msg(location_df)
    print(f'number of tag to remove: {n_bad_labels}')
    print_tags_to_remove(user_msgs)
    print('done')
